from typing import List
from ortools.linear_solver import pywraplp
from draftfast.settings import OptimizerSettings
from draftfast.dke_exceptions import (InvalidBoundsException,
                                      PlayerBanAndLockException)
from draftfast.orm import Player
from draftfast.rules import RuleSet
from draftfast.lineup_constraints import LineupConstraints


class Optimizer(object):
    def __init__(
        self,
        players: List[Player],
        rule_set: RuleSet,
        settings: OptimizerSettings,
        lineup_constraints: LineupConstraints,
        exposure_dict: dict
    ):
        self.solver = pywraplp.Solver(
            'FD',
            pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING
        )
        self.players = players
        self.enumerated_players = list(enumerate(players))
        self.existing_rosters = settings.existing_rosters or []
        self.salary_min = rule_set.salary_min
        self.salary_max = rule_set.salary_max
        self.roster_size = rule_set.roster_size
        self.position_limits = rule_set.position_limits
        self.offensive_positions = rule_set.offensive_positions
        self.defensive_positions = rule_set.defensive_positions
        self.general_position_limits = rule_set.general_position_limits
        self.max_players_per_team = rule_set.max_players_per_team
        self.showdown = rule_set.game_type == 'showdown'
        self.settings = settings
        self.lineup_constraints = lineup_constraints
        self.banned_for_exposure = exposure_dict.get('banned', [])
        self.locked_for_exposure = exposure_dict.get('locked', [])
        self.custom_rules = settings.custom_rules
        self.min_teams = rule_set.min_teams or settings.min_teams
        self.position_per_team_rules = rule_set.position_per_team_rules

        self.player_to_idx_map = {}
        self.name_to_idx_map = {}
        self.variables = []
        self.name_to_idx_map = dict()
        self.player_to_idx_map = dict()

        for idx, player in self.enumerated_players:
            self.variables.append(
                self.solver.IntVar(0, 1, player.solver_id)
            )

            self._add_player_to_idx_maps(player, idx)

            if self._is_locked(player):
                player.lock = True
            if self._is_banned(player):
                player.ban = True
            if self._is_position_locked(player):
                player.position_lock = True
            if self._is_position_banned(player):
                player.position_ban = True

            # TODO: this can only happen because of exposure, but it could be
            # handled better
            if player.lock and player.ban:
                raise PlayerBanAndLockException(player.name)

        self.teams = set([p.team for p in self.players])
        self.objective = self.solver.Objective()
        self.objective.SetMaximization()

    def _add_player_to_idx_maps(self, p: Player, idx: int):
        self.player_to_idx_map[p.solver_id] = idx

        if p.name not in self.name_to_idx_map.keys():
            self.name_to_idx_map[p.name] = set()
        self.name_to_idx_map[p.name].update([idx])

    def _is_locked(self, p: Player) -> bool:
        return self.lineup_constraints.is_locked(p.name) or \
               p.name in self.locked_for_exposure or \
               p.lock

    def _is_banned(self, p: Player) -> bool:
        return self.lineup_constraints.is_banned(p.name) or \
               p.name in self.banned_for_exposure or \
               p.ban

    def _is_position_locked(self, p: Player) -> bool:
        return self.lineup_constraints.is_position_locked(p.solver_id)

    def _is_position_banned(self, p: Player) -> bool:
        return self.lineup_constraints.is_position_banned(p.solver_id)

    def solve(self) -> bool:
        self._set_player_constraints()
        self._set_player_group_constraints()
        self._optimize_on_projected_points()
        self._set_salary_range()
        self._set_roster_size()
        self._set_positions()
        self._set_general_positions()
        self._set_stack()
        self._set_combo()
        self._set_no_duplicate_lineups()
        self._set_min_teams()
        self._set_custom_rules()
        self._set_position_team_constraints()

        if self.offensive_positions and self.defensive_positions \
                and self.settings.no_offense_against_defense or \
                self.showdown and self.settings.no_defense_against_captain:
            self._set_no_opp_defense()

        solution = self.solver.Solve()

        return solution == self.solver.OPTIMAL

    def _set_player_constraints(self):
        multi_constraints = dict()

        for i, p in self.enumerated_players:
            lb = 1 if (p.lock or p.position_lock) else 0
            ub = 0 if (p.ban or p.position_ban) else 1

            if lb > ub:
                raise InvalidBoundsException(
                    'Invalid bounds for {}'.format(
                        p.name
                    )
                )

            if (p.multi_position or self.showdown) and not (
                    p.position_lock or p.position_ban):
                if p.name not in multi_constraints.keys():
                    multi_constraints[p.name] = self.solver.Constraint(lb, ub)
                constraint = multi_constraints[p.name]
            elif (p.multi_position or self.showdown) and p.position_lock:
                if p.name not in multi_constraints.keys():
                    multi_constraints[p.name] = self.solver.Constraint(0, ub)
                multi_constraints[p.name].SetCoefficient(self.variables[i], 1)

                constraint = self.solver.Constraint(lb, ub)
            else:
                constraint = self.solver.Constraint(lb, ub)

            constraint.SetCoefficient(self.variables[i], 1)

    def _set_player_group_constraints(self):
        for group_constraint in self.lineup_constraints:
            if group_constraint.exact:
                lb = ub = group_constraint.exact
            else:
                lb = group_constraint.lb
                ub = group_constraint.ub

            constraint = self.solver.Constraint(lb, ub)
            for name in group_constraint.players:
                for idx in self.name_to_idx_map[name]:
                    constraint.SetCoefficient(self.variables[idx], 1)

    def _optimize_on_projected_points(self):
        for i, player in self.enumerated_players:
            self.objective.SetCoefficient(
                self.variables[i],
                player.proj,
            )

    def _set_salary_range(self):
        salary_cap = self.solver.Constraint(
            self.salary_min,
            self.salary_max,
        )
        for i, player in self.enumerated_players:
            salary_cap.SetCoefficient(
                self.variables[i],
                player.cost
            )

    def _set_roster_size(self):
        size_cap = self.solver.Constraint(
            self.roster_size,
            self.roster_size,
        )

        for variable in self.variables:
            size_cap.SetCoefficient(variable, 1)

    def _set_stack(self):
        if self.settings:
            stacks = self.settings.stacks

            if stacks:
                for stack in stacks:
                    stack_team = stack.team
                    stack_count = stack.count
                    stack_lock_pos = stack.stack_lock_pos
                    stack_lock_eligible_pos = stack.stack_eligible_pos

                    stack_cap = self.solver.Constraint(
                        stack_count,
                        stack_count,
                    )

                    for i, player in self.enumerated_players:
                        if stack_team == player.team:
                            stack_cap.SetCoefficient(
                                self.variables[i],
                                1
                            )

                    self._set_stacking_type(
                        stack_lock_pos,
                        stack_lock_eligible_pos,
                        stack_team,
                        stack.count
                    )

    def _set_stacking_type(
        self,
        stack_lock_pos,
        stack_eligible_pos,
        team,
        count,
    ):
        if self.settings:
            if stack_lock_pos and stack_eligible_pos:
                enumerated_players = self.enumerated_players

                skillplayers_on_team = [
                    self.variables[i] for i, p in enumerated_players
                    if p.team == team and p.pos in stack_eligible_pos
                ]
                locked_on_team = [
                    self.variables[i] for i, p in enumerated_players
                    if p.team == team and p.pos == stack_lock_pos
                ]
                self.solver.Add(
                    self.solver.Sum(skillplayers_on_team) >=
                    self.solver.Sum(locked_on_team)
                )
                self.solver.Add(
                    (self.solver.Sum(skillplayers_on_team)) >=
                    count - 1
                )

    def _set_combo(self):
        if self.settings:
            combo = self.settings.force_combo
            combo_allow_te = self.settings.combo_allow_te

            combo_skill_type = ['WR']
            if combo_allow_te:
                combo_skill_type.append('TE')

            if combo:
                teams = set([p.team for p in self.players])
                enumerated_players = self.enumerated_players

                for team in teams:
                    skillplayers_on_team = [
                        self.variables[i] for i, p in enumerated_players
                        if p.team == team and p.pos in combo_skill_type
                    ]
                    qbs_on_team = [
                        self.variables[i] for i, p in enumerated_players
                        if p.team == team and p.pos == 'QB'
                    ]
                    self.solver.Add(
                        self.solver.Sum(skillplayers_on_team) >=
                        self.solver.Sum(qbs_on_team)
                    )

    def _set_no_opp_defense(self):
        offensive_pos = self.offensive_positions
        defensive_pos = self.defensive_positions

        enumerated_players = self.enumerated_players

        for team in self.teams:
            offensive_against = [
                self.variables[i] for i, p in enumerated_players
                if p.pos in offensive_pos and
                p.is_opposing_team_in_match_up(team)
            ]

            # TODO this is gross for showdown
            defensive = [
                self.variables[i] for i, p in enumerated_players
                if p.team == team and p.pos in defensive_pos or
                self.showdown and p.real_pos in defensive_pos
            ]

            for p in offensive_against:
                for d in defensive:
                    # For each combination of offensive player and their
                    # opposing defense, force no defense given offense (d <= 0)
                    self.solver.Add(d <= 1 - p)

    def _set_position_team_constraints(self):
        if self.position_per_team_rules:
            for team in self.teams:
                for rule in self.position_per_team_rules:
                    position_group_func, max_pos = rule
                    grouped_position_by_team = [
                        self.variables[i] for i, p
                        in self.enumerated_players if p.team == team
                        and position_group_func(p.pos)
                    ]
                    self.solver.Add(
                        max_pos >=
                        self.solver.Sum(grouped_position_by_team)
                    )

    def _set_custom_rules(self):
        if self.custom_rules:
            for rule in self.custom_rules:
                group_a = [
                    self.variables[i] for i, p
                    in self.enumerated_players if rule.group_a(p)
                ]
                group_b = [
                    self.variables[i] for i, p
                    in self.enumerated_players if rule.group_b(p)
                ]
                self.solver.Add(
                    rule.comparison(self.solver.Sum, group_a, group_b)
                )

    def _set_positions(self):
        for position, min_limit, max_limit in self.position_limits:
            position_cap = self.solver.Constraint(
                min_limit,
                max_limit
            )

            for i, player in self.enumerated_players:
                if position == player.pos:
                    position_cap.SetCoefficient(self.variables[i], 1)

    def _set_general_positions(self):
        for general_position, min_limit, max_limit in \
                self.general_position_limits:
            position_cap = self.solver.Constraint(min_limit, max_limit)

            for i, player in self.enumerated_players:
                if general_position == player.mlb_general_position:
                    position_cap.SetCoefficient(
                        self.variables[i],
                        1
                    )
                if general_position == player.nba_general_position:
                    position_cap.SetCoefficient(
                        self.variables[i],
                        1
                    )

    def _set_no_duplicate_lineups(self):
        for roster in self.existing_rosters:
            max_repeats = self.roster_size - 1
            if self.settings.uniques:
                max_repeats = max(
                    self.roster_size - self.settings.uniques,
                    1
                )
            repeated_players = self.solver.Constraint(
                0,
                max_repeats
            )
            for player in roster.sorted_players():
                i = self.player_to_idx_map.get(player.solver_id)
                if i is not None:
                    repeated_players.SetCoefficient(self.variables[i], 1)

    def _set_min_teams(self):
        teams = []
        min_teams = self.min_teams

        if min_teams > 1:
            for team in self.teams:
                if team:
                    team_var = self.solver.IntVar(0, 1, team)
                    teams.append(team_var)
                    players_on_team = [
                        self.variables[i] for i, p
                        in self.enumerated_players if p.team == team
                    ]
                    self.solver.Add(
                        team_var <=
                        self.solver.Sum(players_on_team)
                    )
                    self.solver.Add(
                        self.max_players_per_team >=
                        self.solver.Sum(players_on_team)
                    )

        if len(teams) > 0:
            self.solver.Add(
                self.solver.Sum(teams) >= self.settings.min_teams
            )



import csv
from itertools import islice
from draftfast.pickem import pickem_orm


def map_pids(pid_file):
    player_map = {}
    with open(pid_file, 'r') as f:
        n = 0
        fields = None
        for line in f.readlines():
            n += 1
            if 'TeamAbbrev' in line:  # line with field names was found
                fields = line.split(',')
                break

        f.close()
        f = islice(open(pid_file, 'r'), n, None)
        reader = csv.DictReader(f, fieldnames=fields)
        for line in reader:
            player_map[line['Name']] = line['ID']

    return player_map


def write_to_csv(player_map, roster, writer):
    row = {
        pickem_orm.T1: player_map[roster.T1.name],
        pickem_orm.T2: player_map[roster.T2.name],
        pickem_orm.T3: player_map[roster.T3.name],
        pickem_orm.T4: player_map[roster.T4.name],
        pickem_orm.T5: player_map[roster.T5.name],
        pickem_orm.T6: player_map[roster.T6.name],
    }
    writer.writerow(row)



import os
from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.csv_parse import salary_download
from draftfast.orm import Roster

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/soccer-showdown.csv'.format(CURRENT_DIR)


class Showdown(Roster):
    POSITION_ORDER = {
        'M': 0,
        'F': 1,
        'D': 2,
        'GK': 3,
    }


showdown_limits = [
    ['M', 0, 6],
    ['F', 0, 6],
    ['D', 0, 6],
    ['GK', 0, 6],
]


def test_el_dk():
    soccer_rules = rules.RuleSet(
        site=rules.DRAFT_KINGS,
        league='SOCCER_SHOWDOWN',
        roster_size=6,
        position_limits=showdown_limits,
        salary_max=50_000,
        general_position_limits=[],
    )
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=soccer_rules,
        player_pool=player_pool,
        verbose=True,
        roster_gen=Showdown,
    )
    ntools.assert_not_equal(roster, None)



import os
from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.csv_parse import salary_download
from draftfast.settings import OptimizerSettings, Stack

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/dk-nhl-salaries.csv'.format(CURRENT_DIR)


def test_nhl_dk():
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
        ruleset=rules.DK_NHL_RULE_SET,
    )
    roster = run(
        rule_set=rules.DK_NHL_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )
    ntools.assert_not_equal(roster, None)


def test_triple_stack():
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
        ruleset=rules.DK_NHL_RULE_SET,
    )
    roster = run(
        rule_set=rules.DK_NHL_RULE_SET,
        player_pool=player_pool,
        verbose=True,
        optimizer_settings=OptimizerSettings(
            stacks=[
                Stack(team='TOR', count=3),
                Stack(team='COL', count=3),
                Stack(team='VAN', count=2),
            ]
        )
    )
    players = roster.sorted_players()

    phi_players = [x for x in players if x.team == 'TOR']
    fla_players = [x for x in players if x.team == 'COL']
    nsh_players = [x for x in players if x.team == 'VAN']
    ntools.assert_equal(len(phi_players), 3)
    ntools.assert_equal(len(fla_players), 3)
    ntools.assert_equal(len(nsh_players), 2)



from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.orm import Player

GOLFER = 'G'


def test_golf_dk():
    player_pool = [
        Player(name='G1', cost=5500, proj=55, pos=GOLFER),
        Player(name='G2', cost=5600, proj=55, pos=GOLFER),
        Player(name='G3', cost=5700, proj=55, pos=GOLFER),
        Player(name='G4', cost=5800, proj=55, pos=GOLFER),
        Player(name='G5', cost=5800, proj=55, pos=GOLFER),
        Player(name='G6', cost=5900, proj=55, pos=GOLFER),
        Player(name='G7', cost=10000, proj=155, pos=GOLFER),
    ]
    roster = run(
        rule_set=rules.DK_PGA_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )
    ntools.assert_not_equal(roster, None)



import os
from draftfast.rules import DRAFT_KINGS, FAN_DUEL

upload_file = '{}/data/current-upload.csv'.format(os.getcwd())


def write_to_csv(writer, player_map, roster, game=DRAFT_KINGS,
                 league='NBA'):
    players = roster.sorted_players()

    ordered_possible = []
    if game == DRAFT_KINGS and league == 'EL':
        ordered_possible = [
            _on_position(players, ['G']),
            _on_position(players, ['G']),
            _on_position(players, ['F']),
            _on_position(players, ['F']),
            _on_position(players, ['F']),
            players
        ]
    elif game == DRAFT_KINGS and league == 'NFL':
        ordered_possible = [
            _on_position(players, ['QB']),
            _on_position(players, ['RB']),
            _on_position(players, ['RB']),
            _on_position(players, ['WR']),
            _on_position(players, ['WR']),
            _on_position(players, ['WR']),
            _on_position(players, ['TE']),

            # NFL Flex
            _on_position(players, ['TE', 'WR', 'RB']),
            _on_position(players, ['DST']),
        ]
    elif game == DRAFT_KINGS and league == 'XFL':
        ordered_possible = [
            _on_position(players, ['QB']),
            _on_position(players, ['RB']),
            _on_position(players, ['WR']),
            _on_position(players, ['WR']),
            _on_position(players, ['WR', 'RB']),
            _on_position(players, ['WR', 'RB']),
            _on_position(players, ['DST']),
        ]
    elif game == DRAFT_KINGS and league == 'SOCCER':
        ordered_possible = [
            _on_position(players, ['F']),
            _on_position(players, ['F']),
            _on_position(players, ['M']),
            _on_position(players, ['M']),
            _on_position(players, ['D']),
            _on_position(players, ['D']),
            _on_position(players, ['GK']),
            players
        ]
    elif game == DRAFT_KINGS and league == 'NHL':
        ordered_possible = [
            _on_position(players, ['C']),
            _on_position(players, ['C']),
            _on_position(players, ['W']),
            _on_position(players, ['W']),
            _on_position(players, ['W']),
            _on_position(players, ['D']),
            _on_position(players, ['D']),
            _on_position(players, ['G']),
            players
        ]
    elif game == DRAFT_KINGS:
        ordered_possible = [
            _on_position(players, ['PG']),
            _on_position(players, ['SG']),
            _on_position(players, ['SF']),
            _on_position(players, ['PF']),
            _on_position(players, ['C']),
            _on_position(players, ['SG', 'PG']),
            _on_position(players, ['SF', 'PF']),
            players
        ]
    elif game == FAN_DUEL:
        ordered_possible = [
            _on_position(players, ['PG']),
            _on_position(players, ['PG']),
            _on_position(players, ['SG']),
            _on_position(players, ['SG']),
            _on_position(players, ['SF']),
            _on_position(players, ['SF']),
            _on_position(players, ['PF']),
            _on_position(players, ['PF']),
            _on_position(players, ['C']),
        ]

    ordered_lineup = []
    counter = 0
    for ps in ordered_possible:
        counter += 1
        not_used_ps = [
            p for p in ps
            if p not in ordered_lineup
        ]
        ordered_lineup.append(not_used_ps[0])

    writer.writerow([
        p.get_player_id(player_map)
        for p in ordered_lineup
    ])


def _on_position(players, possible):
    return [p for p in players if p.pos in possible]



import argparse
import os

OPTIMIZE_COMMAND_LINE = [
    ['-use_averages', 'use averages as projections', False],
    ['-v_avg', 'projections must be within points v avg', 100000],
    ['-min_avg', 'player must exceed average points', 0],
    ['-lp', 'lowest acceptable projection', 0],
    ['-home', 'only select players playing at home', None],
    ['-locked_teams', 'player must be on specified teams', None],
    ['-banned_teams', 'player must not be on specified teams', None],
    ['-source', 'data source to use', 'nfl_rotogrinders'],
    ['-pids', 'player id file (create upload file)', None],
    ['-keep_pids', 'Keep current upload file', None],
    ['-salary_file', 'file location for salaries',
     os.getcwd() + '/data/current-salaries.csv'],
    ['-projection_file', 'file location for projections',
     os.getcwd() + '/data/current-projections.csv'],
    ['-randomize_projections', 'use random projection factor', None],
]


MULTIPLE_ARGS_COMMAND = [
    '-teams',
]

PARSER = argparse.ArgumentParser()


def get_args():
    for opt in OPTIMIZE_COMMAND_LINE:
        nargs = '?'
        if opt[0] in MULTIPLE_ARGS_COMMAND:
            nargs = '+'
        PARSER.add_argument(
            opt[0],
            nargs=nargs,
            help=opt[1],
            default=opt[2]
        )
    return PARSER.parse_args()



import random
from nose import tools as ntools
from draftfast.player_pool import filter_pool
from draftfast.orm import Player
from draftfast.settings import PlayerPoolSettings


p_a, p_b, p_c = [
    Player(name='A1', cost=5500, proj=20, pos='PG'),
    Player(name='A2', cost=7000, proj=30, pos='PG'),
    Player(name='A3', cost=10000, proj=55, pos='PG'),
]
mock_player_pool = [p_a, p_b, p_c]


def test_no_settings():
    pool = filter_pool(mock_player_pool, PlayerPoolSettings())
    ntools.assert_equals(pool, mock_player_pool)


def test_respects_min_proj():
    pool = filter_pool(mock_player_pool, PlayerPoolSettings(min_proj=50))
    ntools.assert_equals(pool, [p_c])


def test_respects_max_proj():
    pool = filter_pool(mock_player_pool, PlayerPoolSettings(max_proj=50))
    ntools.assert_equals(pool, [p_a, p_b])


def test_respects_min_cost():
    pool = filter_pool(mock_player_pool, PlayerPoolSettings(min_salary=7001))
    ntools.assert_equals(pool, [p_c])


def test_respects_max_cost():
    pool = filter_pool(mock_player_pool, PlayerPoolSettings(max_salary=7001))
    ntools.assert_equals(pool, [p_a, p_b])


def test_randomize():
    random.seed(1)
    pool = filter_pool(mock_player_pool, PlayerPoolSettings(
        randomize=0.1,
    ))
    ntools.assert_equals(
        pool[0].proj,
        18.537456976449604
    )



import os
from copy import deepcopy
from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.orm import Player
from draftfast.csv_parse import salary_download
from draftfast.settings import OptimizerSettings, Stack
from draftfast.lineup_constraints import LineupConstraints

mock_nba_pool = [
    Player(name='A1', cost=5500, proj=40, pos='PG'),
    Player(name='A2', cost=5500, proj=41, pos='PG'),
    Player(name='A11', cost=5500, proj=50, pos='PG'),
    Player(name='A3', cost=5500, proj=42, pos='SG'),
    Player(name='A4', cost=5500, proj=43, pos='SG'),
    Player(name='A5', cost=5500, proj=44, pos='SF'),
    Player(name='A6', cost=5500, proj=45, pos='SF'),
    Player(name='A7', cost=5500, proj=46, pos='PF'),
    Player(name='A8', cost=5500, proj=47, pos='PF'),
    Player(name='A9', cost=5500, proj=48, pos='C'),
    Player(name='A10', cost=5500, proj=49, pos='C'),
]

mock_nfl_pool = [
    Player(name='A1', cost=5500, proj=40, pos='QB'),
    Player(name='A2', cost=5500, proj=41, pos='QB'),
    Player(name='A11', cost=5500, proj=50, pos='WR'),
    Player(name='A3', cost=5500, proj=42, pos='WR'),
    Player(name='A4', cost=5500, proj=43, pos='WR'),
    Player(name='A5', cost=5500, proj=44, pos='WR'),
    Player(name='A6', cost=5500, proj=45, pos='RB'),
    Player(name='A7', cost=5500, proj=46, pos='RB'),
    Player(name='A8', cost=5500, proj=47, pos='RB'),
    Player(name='A9', cost=5500, proj=48, pos='TE'),
    Player(name='A10', cost=5500, proj=49, pos='TE'),
    Player(name='A12', cost=5500, proj=51, pos='DST'),
    Player(name='A13', cost=5500, proj=52, pos='DST'),
]


CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/dk-nfl-salaries.csv'.format(CURRENT_DIR)
fd_nfl_salary_file = '{}/data/fd-nfl-salaries.csv'.format(CURRENT_DIR)
projection_file = '{}/data/dk-nfl-projections.csv'.format(CURRENT_DIR)


def test_nba_dk():
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
    )
    ntools.assert_not_equal(roster, None)


def test_nba_dk_with_csv():
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
    )
    ntools.assert_not_equal(roster, None)


def test_nba_fd():
    roster = run(
        rule_set=rules.FD_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True
    )
    ntools.assert_not_equal(roster, None)


def test_nfl_dk_mock():
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=mock_nfl_pool,
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 420.0)


def test_nfl_dk():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,

    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 124.30)


def test_nfl_fd():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=fd_nfl_salary_file,
        game=rules.FAN_DUEL,
    )
    roster = run(
        rule_set=rules.FD_NFL_RULE_SET,
        player_pool=players,
        verbose=True
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 155.0172712846236)


def test_multi_position():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        constraints=LineupConstraints(
            locked=['Eli Manning'],
        ),
        verbose=True
    )
    ntools.assert_not_equal(roster, None)
    multi_pos = [p for p in roster.players if p.name == 'Eli Manning']
    ntools.assert_equal(len(multi_pos), 1)
    ntools.assert_equal(multi_pos[0].pos, 'TE')


def test_multi_roster_nfl():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True
    )
    second_roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            existing_rosters=[roster],
        ),
        verbose=True
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_not_equal(second_roster, None)
    ntools.assert_false(roster == second_roster)


def test_multi_roster_nba():
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
    )
    second_roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        optimizer_settings=OptimizerSettings(
            existing_rosters=[roster],
        ),
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_not_equal(second_roster, None)
    ntools.assert_false(roster == second_roster)


def test_uniques_nba():
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
    )
    second_roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        optimizer_settings=OptimizerSettings(
            existing_rosters=[roster],
        ),
    )
    third_roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        optimizer_settings=OptimizerSettings(
            existing_rosters=[roster],
            uniques=2,
        ),
    )

    players = roster.sorted_players()
    second_players = second_roster.sorted_players()
    third_players = third_roster.sorted_players()
    crossover_a = list(set(players).intersection(second_players))
    crossover_b = list(set(players).intersection(third_players))
    ntools.assert_equal(
        len(crossover_a),
        rules.DK_NBA_RULE_SET.roster_size - 1
    )
    ntools.assert_equal(
        len(crossover_b),
        rules.DK_NBA_RULE_SET.roster_size - 2
    )


def test_respect_lock():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True,
        constraints=LineupConstraints(
            locked=['Andrew Luck'],
        ),
    )
    qb = roster.sorted_players()[0]
    ntools.assert_equal(qb.pos, 'QB')
    ntools.assert_equal(qb.name, 'Andrew Luck')


def test_respect_ban():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True,
        constraints=LineupConstraints(
            banned=['Eli Manning'],
        ),
    )
    for player in roster.sorted_players():
        ntools.assert_not_equal(player.name, 'Eli Manning')


def test_respect_group1():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    grouped_players = ('DeAndre Hopkins', 'Amari Cooper', 'Sammy Watkins')

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True,
        constraints=LineupConstraints(
            groups=[
                [grouped_players, 2]
            ],
        ),
    )

    group_count = len([
        x for x in roster.sorted_players() if x.name in grouped_players
    ])
    ntools.assert_equal(group_count, 2)


def test_respect_group2():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    grouped_players = (
        'Ryan Fitzpatrick',
        'Lamar Miller',
        'DeAndre Hopkins',
        'Amari Cooper',
        'Sammy Watkins'
    )

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True,
        constraints=LineupConstraints(
            groups=[
                [grouped_players, (2, 3)]
            ],
        ),
    )

    group_count = len([
        x for x in roster.sorted_players() if x.name in grouped_players
    ])
    ntools.assert_true(group_count >= 2 and group_count <= 3)


def test_respect_group3():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    grouped_players = (
        'Ryan Fitzpatrick',
        'Lamar Miller',
        'DeAndre Hopkins',
        'Amari Cooper',
        'Sammy Watkins'
    )

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        verbose=True,
        constraints=LineupConstraints(
            groups=[
                [grouped_players, 1]
            ],
        ),
    )

    group_count = len([
        x for x in roster.sorted_players() if x.name in grouped_players
    ])
    ntools.assert_true(group_count == 1)


def test_stack():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            stacks=[
                Stack(team='NE', count=5)
            ]
        ),
        verbose=True,
    )
    ne_players_count = len([
        p for p in roster.sorted_players()
        if p.team == 'NE'
    ])
    ntools.assert_equals(5, ne_players_count)
    ne_def = len([
        p for p in roster.sorted_players()
        if p.team == 'NE' and p.pos == 'DST'
    ])
    ntools.assert_equals(ne_def, 1)


def test_custom_stack():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            stacks=[
                Stack(
                    team='NE',
                    count=5,
                    stack_lock_pos=['QB'],
                    stack_eligible_pos=['WR'],
                )
            ]
        ),
        verbose=True,
    )
    ne_players_count = len([
        p for p in roster.sorted_players()
        if p.team == 'NE'
    ])
    ntools.assert_equals(5, ne_players_count)
    ne_def = len([
        p for p in roster.sorted_players()
        if p.team == 'NE' and p.pos == 'DST'
    ])
    ntools.assert_equals(ne_def, 0)
    wides = len([
        p for p in roster.sorted_players()
        if p.team == 'NE' and p.pos == 'WR'
    ])
    ntools.assert_equals(wides, 4)


def test_force_combo():
    # no combo
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            stacks=[
                Stack(team='NE', count=5)
            ]
        ),
        constraints=LineupConstraints(
            locked=['Sam Bradford'],
        ),
    )
    qb = roster.sorted_players()[0]
    team_count = len([
        x for x in roster.sorted_players()
        if x.team == qb.team
    ])
    ntools.assert_equals(team_count, 1)

    # QB/WR combo
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            force_combo=True,
        ),
        constraints=LineupConstraints(
            banned=['Ryan Fitzpatrick']
        ),
        verbose=True,
    )
    qb = roster.sorted_players()[0]
    ntools.assert_equal(qb.pos, 'QB')

    wr_team_count = len([
        x for x in roster.sorted_players()
        if x.team == qb.team and x.pos == 'WR'
    ])
    ntools.assert_equals(wr_team_count, 1)

    te_team_count = len([
        x for x in roster.sorted_players()
        if x.team == qb.team and x.pos == 'TE'
    ])
    ntools.assert_equals(te_team_count, 0)


def test_te_combo():
    # use lock and ban to make a non-globally optimal QB/TE combo optimal
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            force_combo=True,
            combo_allow_te=True,
        ),
        constraints=LineupConstraints(
            banned=['Kellen Davis'],
            locked=['Philip Rivers'],
        ),
        verbose=True,
    )
    qb = roster.sorted_players()[0]
    ntools.assert_equal(qb.pos, 'QB')
    team_count = len([
        x for x in roster.sorted_players()
        if x.team == qb.team and x.pos == 'TE'
    ])
    ntools.assert_equals(team_count, 1)

    # make sure WR/QB still works
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            force_combo=True,
            combo_allow_te=True,
        ),
        constraints=LineupConstraints(
            locked=['Andrew Luck'],
        ),
        verbose=True,
    )
    qb = roster.sorted_players()[0]
    ntools.assert_equal(qb.pos, 'QB')
    team_count = len([
        x for x in roster.sorted_players()
        if x.team == qb.team and x.pos == 'WR'
    ])
    ntools.assert_equals(team_count, 1)


def test_impossible_constraints():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            stacks=[
                Stack(team='NE', count=100)
            ],
            no_offense_against_defense=True
        ),
        constraints=LineupConstraints(
            banned=['Sammy Watkins', 'Kellen Davis'],
            locked=['Spencer Ware'],
            groups=[
                [('Philip Rivers', 'Sam Bradford', 'Andrew Luck'), 2],
                [('Saints', 'Alshon Jeffery', 'Lamar Miller'), (1, 3)]
            ]
        ),
        verbose=True
    )

    ntools.assert_equal(roster, None)


def test_multi_position_group_constraint():
    players = [
        Player(name='A', cost=5500, proj=400, pos='QB',
               possible_positions='QB/WR', multi_position=True),
        Player(name='A', cost=5500, proj=400, pos='WR',
               possible_positions='QB/WR', multi_position=True),
        Player(name='B', cost=5500, proj=41, pos='QB'),
        Player(name='C', cost=5500, proj=500, pos='WR',
               possible_positions='RB/WR', multi_position=True),
        Player(name='C', cost=5500, proj=500, pos='RB',
               possible_positions='RB/WR', multi_position=True),
        Player(name='D', cost=5500, proj=42, pos='WR'),
        Player(name='E', cost=5500, proj=43, pos='WR'),
        Player(name='F', cost=5500, proj=44, pos='WR'),
        Player(name='G', cost=5500, proj=45, pos='RB'),
        Player(name='H', cost=5500, proj=46, pos='RB'),
        Player(name='I', cost=5500, proj=47, pos='RB'),
        Player(name='J', cost=5500, proj=480, pos='TE',
               possible_positions='TE/WR', multi_position=True),
        Player(name='J', cost=5500, proj=480, pos='WR',
               possible_positions='TE/WR', multi_position=True),
        Player(name='K', cost=5500, proj=49, pos='TE'),
        Player(name='L', cost=5500, proj=51, pos='DST'),
        Player(name='M', cost=5500, proj=52, pos='DST'),
    ]

    grouped_players = ('A', 'C', 'J')

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        constraints=LineupConstraints(
            groups=[
                [grouped_players, 2],
            ]
        ),
        verbose=True
    )

    group_count = len([
        x for x in roster.sorted_players() if x.name in grouped_players
    ])
    ntools.assert_equal(group_count, 2)
    ntools.assert_equal(roster.projected(), 1304)


def test_multi_position_group_constraint2():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    grouped_players = ('Eli Manning', 'Dez Bryant', 'Geno Smith')

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        constraints=LineupConstraints(
            groups=[
                [grouped_players, 2],
            ]
        ),
        verbose=True
    )

    group_count = len([
        x for x in roster.sorted_players() if x.name in grouped_players
    ])
    ntools.assert_equal(group_count, 2)
    ntools.assert_equal(roster.projected(), 120.89999999999999)


def test_no_opposing_def_dk_nfl_mock():
    mock_pool = [
        Player(name='A1', cost=5500, proj=40, pos='QB', team='X',
               matchup='X@Y'),
        Player(name='A2', cost=5500, proj=41, pos='QB', team='Y',
               matchup='X@Y'),
        Player(name='A11', cost=5500, proj=50, pos='WR', team='X',
               matchup='X@Y'),
        Player(name='A3', cost=5500, proj=42, pos='WR', team='Y',
               matchup='X@Y'),
        Player(name='A4', cost=5500, proj=43, pos='WR', team='X',
               matchup='X@Y'),
        Player(name='A5', cost=5500, proj=44, pos='WR', team='Y',
               matchup='X@Y'),
        Player(name='A111', cost=5500, proj=50, pos='WR', team='X',
               matchup='X@Y'),
        Player(name='A31', cost=5500, proj=42, pos='WR', team='Y',
               matchup='X@Y'),
        Player(name='A41', cost=5500, proj=62, pos='WR', team='X',
               matchup='X@Y'),
        Player(name='A51', cost=5500, proj=63, pos='WR', team='Y',
               matchup='X@Y'),
        Player(name='A6', cost=5500, proj=45, pos='RB', team='X',
               matchup='X@Y'),
        Player(name='A7', cost=5500, proj=46, pos='RB', team='Y',
               matchup='X@Y'),
        Player(name='A8', cost=5500, proj=47, pos='RB', team='X',
               matchup='X@Y'),
        Player(name='A71', cost=5500, proj=46, pos='RB', team='Y',
               matchup='X@Y'),
        Player(name='A81', cost=5500, proj=47, pos='RB', team='X',
               matchup='X@Y'),
        Player(name='A711', cost=5500, proj=46, pos='RB', team='Y',
               matchup='X@Y'),
        Player(name='A9', cost=5500, proj=48, pos='TE', team='X',
               matchup='X@Y'),
        Player(name='A10', cost=5500, proj=49, pos='TE', team='Y',
               matchup='X@Y'),
        Player(name='A12', cost=5500, proj=51, pos='DST', team='X',
               matchup='X@Y'),
        Player(name='A13', cost=5500, proj=500, pos='DST', team='Y',
               matchup='X@Y'),
    ]

    # mock pool is constructed such that optimal lineup has qb opposing dst
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=mock_pool,
        verbose=True
    )

    ntools.assert_equal(roster.projected(), 909)
    qb_team = roster.sorted_players()[0].team
    dst_team = roster.sorted_players()[-1].team
    ntools.assert_equal(qb_team, dst_team)

    # this will fail to produce a roster because we only have 2 teams (X and Y)
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=mock_pool,
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        verbose=True
    )

    ntools.assert_equal(roster, None)

    # relax min teams
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=mock_pool,
        optimizer_settings=OptimizerSettings(
            min_teams=1,
            no_offense_against_defense=True,
        ),
        verbose=True
    )

    ntools.assert_equal(roster.projected(), 877)
    ntools.assert_equal(len(set([p.team for p in roster.players])), 1)

    # add a player from a third team, min 2 teams
    mock_pool.append(Player(
        name='B2', cost=5500, proj=70, pos='QB', team='Q', matchup='Q@Z'
    )
    )
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=deepcopy(mock_pool),
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        verbose=True
    )

    for p in roster.players:
        if p.pos in rules.DK_NFL_RULE_SET.offensive_positions:
            ntools.assert_not_equal(p.team, 'X')


def test_no_opposing_def_dk_nfl():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )

    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        constraints=LineupConstraints(
            locked=['Bengals']
        ),
        verbose=True
    )

    for p in roster.players:
        if p.pos in rules.DK_NFL_RULE_SET.offensive_positions:
            ntools.assert_not_equal(p.team, 'CIN')

    # force impossible lineup
    roster = run(
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        constraints=LineupConstraints(
            locked=['Bengals', 'Ryan Fitzpatrick']
        ),
        verbose=True
    )

    ntools.assert_equal(roster, None)


def test_no_opposing_def_fd_nfl():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=fd_nfl_salary_file,
        game=rules.FAN_DUEL,
    )
    roster = run(
        rule_set=rules.FD_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        constraints=LineupConstraints(
            locked=['Jacksonville Jaguars']
        ),
        verbose=True
    )

    ntools.assert_not_equal(roster, None)

    for p in roster.players:
        if p.pos in rules.DK_NFL_RULE_SET.offensive_positions:
            ntools.assert_not_equal(p.team, 'IND')

    # force impossible lineup
    roster = run(
        rule_set=rules.FD_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        constraints=LineupConstraints(
            locked=['Eric Ebron', 'Jacksonville Jaguars']
        ),
        verbose=True
    )

    ntools.assert_equal(roster, None)


def test_no_mutate_side_effect():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=fd_nfl_salary_file,
        game=rules.FAN_DUEL,
    )
    run(
        rule_set=rules.FD_NFL_RULE_SET,
        player_pool=players,
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True
        ),
        constraints=LineupConstraints(
            locked=['Tom Brady']
        ),
        verbose=True
    )
    brady = next((p for p in players if p.name == 'Tom Brady'))
    ntools.assert_equal(brady.lock, False)



import csv
from itertools import islice
from .upload import (
    write_to_csv,
)
from draftfast.rules import DRAFT_KINGS, FAN_DUEL
from draftfast.pickem import pickem_orm, pickem_upload
from draftfast import dke_exceptions as dke

NAME_MAP = {
    DRAFT_KINGS: {
        'start': 'TeamAbbrev',
        'name': 'Name',
        'position': 'Position',
        'id': 'ID',
    },
    FAN_DUEL: {
        'start': '"Nickname"',
        'name': '"Nickname"',
        'position': '"Position"',
        'id': '"Player ID + Player Name"',
    },
}


def map_pids(pid_file, encoding, errors, game=DRAFT_KINGS):
    start = NAME_MAP.get(game).get('start')
    name = NAME_MAP.get(game).get('name')
    position = NAME_MAP.get(game).get('position')
    p_id = NAME_MAP.get(game).get('id')

    player_map = {}
    with open(pid_file, 'r') as f:
        n = 0
        fields = None
        for line in f.readlines():
            n += 1
            if start in line:  # line with field names was found
                fields = line.split(',')
                break

        if not fields:
            raise dke.InvalidCSVUploadFileException(
                "Check that you're using the DK CSV upload template, " +
                "which can be found at " +
                "https://www.draftkings.com/lineup/upload.")

        f.close()
        f = islice(open(pid_file, 'r',
                        encoding=encoding, errors=errors), n, None)
        reader = csv.DictReader(f, fieldnames=fields)
        for line in reader:
            # DraftKings adds spaces to DST for NFL
            if 'DST' in line[position]:
                line[name] = line[name].strip()
                line[position] = line[position].strip()

            player_map[line[name] + " " + line[position]] = line[p_id]

    return player_map


class CSVUploader(object):

    def __init__(self, pid_file, upload_file='./upload.csv',
                 encoding='utf-8', errors='replace'):
        self.upload_file = upload_file
        self.encoding = encoding
        self.errors = errors
        self.pid_map = self._map_pids(pid_file)

    def _map_pids(self, pid_file):
        raise NotImplementedError('You must implement _map_pids')


class DraftKingsUploader(CSVUploader):
    def write_rosters(self, rosters):
        with open(self.upload_file, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self.HEADERS)
            for roster in rosters:
                write_to_csv(
                    writer=writer,
                    roster=roster,
                    player_map=self.pid_map,
                    league=self.LEAGUE,
                )

    def _map_pids(self, pid_file):
        return map_pids(
            pid_file,
            game=DRAFT_KINGS,
            encoding=self.encoding,
            errors=self.errors,
        )


class DraftKingsNBAUploader(DraftKingsUploader):
    LEAGUE = 'NBA'
    HEADERS = [
        'PG', 'SG', 'SF',
        'PF', 'C', 'G', 'F', 'UTIL'
    ]


class DraftKingsELUploader(DraftKingsUploader):
    LEAGUE = 'EL'
    HEADERS = [
        'G', 'G', 'F', 'F', 'F', 'UTIL',
    ]


class DraftKingsSoccerUploader(DraftKingsUploader):
    LEAGUE = 'SOCCER'
    HEADERS = [
        'F', 'F', 'M', 'M', 'D', 'D', 'GK', 'UTIL',
    ]


class DraftKingsNHLUploader(DraftKingsUploader):
    LEAGUE = 'NHL'
    HEADERS = [
        'C', 'C', 'W', 'W', 'W', 'D',
        'D', 'G', 'UTIL',
    ]


class DraftKingsNFLUploader(DraftKingsUploader):
    LEAGUE = 'NFL'
    HEADERS = [
        'QB', 'RB', 'RB',
        'WR', 'WR', 'WR',
        'TE', 'FLEX', 'DST'
    ]


class DraftKingsXFLUploader(DraftKingsUploader):
    LEAGUE = 'XFL'
    HEADERS = [
        'QB', 'RB',
        'WR', 'WR',
        'FLEX', 'FLEX',
        'DST',
    ]


class FanDuelNBAUploader(CSVUploader):
    HEADERS = [
        'PG', 'PG', 'SG', 'SG', 'SF',
        'SF', 'PF', 'PF', 'C',
    ]

    def write_rosters(self, rosters):
        with open(self.upload_file, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self.HEADERS)
            for roster in rosters:
                write_to_csv(
                    writer=writer,
                    roster=roster,
                    player_map=self.pid_map,
                    game=FAN_DUEL,
                )

    def _map_pids(self, pid_file):
        return map_pids(
            pid_file,
            game=FAN_DUEL,
            encoding=self.encoding,
            errors=self.errors,
        )


class FanDuelNFLUploader(CSVUploader):
    pass


class DraftKingsNBAPickemUploader(CSVUploader):
    def write_rosters(self, rosters):
        with open(self.upload_file, 'w') as f:
            writer = csv.DictWriter(f, fieldnames=pickem_orm.TIERS)
            writer.writeheader()
            for roster in rosters:
                pickem_upload.write_to_csv(
                    writer=writer,
                    roster=roster,
                    player_map=self.pid_map,
                )

    def _map_pids(self, pid_file):
        return pickem_upload.map_pids(pid_file)


class DraftKingsCaptainShowdownUploader(DraftKingsUploader):
    HEADERS = [
        'CPT', 'UTIL', 'UTIL', 'UTIL', 'UTIL', 'UTIL'
    ]

    def write_rosters(self, rosters):
        with open(self.upload_file, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self.HEADERS)
            for roster in rosters:
                writer.writerow([
                    p.get_player_id(self.pid_map)
                    for p in roster.sorted_players()
                ])



from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.orm import Player
from draftfast.settings import OptimizerSettings
from draftfast.showdown.orm import ShowdownPlayer


def _build_mock_player_pool():
    player_pool = [
        Player(name='A1', cost=5500, proj=100, pos='QB',
               team='X', matchup='X@Y'),
        Player(name='A2', cost=5500, proj=41, pos='QB',
               team='X', matchup='X@Y'),
        Player(name='A11', cost=5500, proj=50, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A3', cost=5500, proj=42, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A4', cost=5500, proj=43, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A5', cost=5500, proj=44, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A6', cost=5500, proj=45, pos='RB',
               team='X', matchup='X@Y'),

        # Test that max players per team works. Everyone
        # on Y is projected for 1 point, under normal
        # opto should never be picked.
        Player(name='A7', cost=5500, proj=1, pos='RB',
               team='Y', matchup='X@Y'),
        Player(name='A8', cost=5500, proj=1, pos='RB',
               team='Y', matchup='X@Y'),
        Player(name='A9', cost=5500, proj=1, pos='TE',
               team='Y', matchup='X@Y'),
    ]

    def capt_boost(p):
        return Player(
            name=p.name,
            team=p.team,
            matchup=p.matchup,
            pos=p.pos,
            cost=p.cost * 1.5,
            proj=p.proj * 1.5,
        )

    captain_pool = [capt_boost(p) for p in player_pool]

    mock_dk_pool = []
    for p in player_pool:
        mock_dk_pool.append(ShowdownPlayer(p))
    for p in captain_pool:
        mock_dk_pool.append(ShowdownPlayer(p, captain=True))

    return mock_dk_pool


def test_csgo_mock():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        # CSGO limits three per team
        rule_set=rules.DK_CSGO_SHOWDOWN,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
        ),
        verbose=True
    )

    ntools.assert_not_equal(roster, None)
    players = roster.players
    ntools.assert_equal(len([
        x for x in players if x.team == 'Y'
    ]), 3)



import os
from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.csv_parse import salary_download

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/dk-euro-league-salaries.csv'.format(CURRENT_DIR)


def test_el_dk():
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_EURO_LEAGUE_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )
    ntools.assert_not_equal(roster, None)



class PlayerPoolSettings(object):

    def __init__(self, min_proj=None, max_proj=None,
                 min_avg=None, max_avg=None, min_salary=None,
                 max_salary=None, randomize=None):
        self.min_proj = min_proj
        self.max_proj = max_proj
        self.min_avg = min_avg
        self.max_avg = max_avg
        self.min_salary = min_salary
        self.max_salary = max_salary
        self.randomize = randomize

    def exist(self):
        return str(self) != 'None'

    # TODO: format this like a proper repr(), i.e. <PlayerPoolSettings: ...>
    def __repr__(self):
        if not str(self):
            return '<PlayerPoolSettings: None>'
        else:
            return str(self)

    def __str__(self):
        lines = []
        if self.min_proj:
            lines.append('Min projection: {}'.format(self.min_proj))
        if self.max_proj:
            lines.append('Max projection: {}'.format(self.min_proj))
        if self.min_avg:
            lines.append('Min average: {}'.format(self.min_proj))
        if self.max_avg:
            lines.append('Max average: {}'.format(self.min_proj))
        if self.min_salary:
            lines.append('Min salary: {}'.format(self.min_proj))
        if self.max_salary:
            lines.append('Max salary: {}'.format(self.min_proj))
        if self.randomize:
            lines.append('Randomization factor: {}'.format(self.min_proj))

        if len(lines):
            return '\n'.join(lines)
        else:
            return 'None'


def default_comparison(solver_sum, g_a, g_b):
    return solver_sum(g_a) + 1 <= solver_sum(g_b)


class CustomRule(object):
    def __init__(self, group_a, group_b, comparison=None):
        self.group_a = group_a
        self.group_b = group_b
        self.comparison = comparison or default_comparison

    def __repr__(self):
        import inspect as i
        return f"{i.getsource(self.group_a)}" \
               f"{i.getsource(self.group_b)}" \
               f"{i.getsource(self.comparison)}"


class OptimizerSettings(object):

    def __init__(
        self,
        stacks=None,
        existing_rosters=None,
        force_combo=None,
        combo_allow_te=None,
        uniques=None,
        no_offense_against_defense=False,
        no_defense_against_captain=False,
        showdown_teams=None,
        min_teams=2,
        custom_rules=None,
    ):
        self.stacks = stacks
        self.existing_rosters = existing_rosters or []
        self.force_combo = force_combo
        self.combo_allow_te = combo_allow_te
        self.uniques = uniques
        self.no_offense_against_defense = no_offense_against_defense
        self.no_defense_against_captain = no_defense_against_captain
        self.showdown_teams = showdown_teams
        self.min_teams = min_teams
        self.custom_rules = custom_rules

    # TODO: format this like a proper repr(), i.e. <OptimizerSettings: ...>
    def __repr__(self):
        if not str(self):
            return '<OptimizerSettings: None>'
        else:
            return str(self)

    def __str__(self):
        lines = []
        if self.stacks and len(self.stacks):
            lines.append('Stacks: {}'.format(
                        [(x.team, x.count) for x in self.stacks]
                    )
                )
        if self.min_teams:
            lines.append('Min teams: {}'.format(self.min_teams))
        if self.no_offense_against_defense:
            lines.append('No offense against D: {}'.format(
                    self.no_offense_against_defense
                )
            )
        if self.custom_rules:
            lines.append('Custom rules: {}'.format(
                self.custom_rules
            ))

        if len(lines):
            return '\n'.join(lines)
        else:
            return 'None'


class UploadSettings(object):

    def __init__(self, pid_file, upload_file,
                 rule_set, rosters):
        self.pid_file = pid_file
        self.upload_file = upload_file
        self.rule_set = rule_set
        self.rosters = rosters


class Stack(object):
    def __init__(
        self,
        team: str,
        count: int,
        stack_lock_pos=None,
        stack_eligible_pos=None
    ):
        self.team = team
        self.count = count
        self.stack_lock_pos = stack_lock_pos
        self.stack_eligible_pos = stack_eligible_pos



from draftfast.orm import Player
from copy import deepcopy


class ShowdownPlayer(Player):
    def __init__(self, player: Player, captain: bool = False):
        for k, v in player.__dict__.items():
            if hasattr(self, k) or k.startswith('__'):
                continue
            setattr(self, k, deepcopy(v))

        if captain:
            self.real_pos = self.pos
            self.pos = 'CPT'
            self.captain = True
        else:
            self.real_pos = self.pos
            self.pos = 'FLEX'
            self.captain = False

    @property
    def formatted_position(self):
        return '{} ({})'.format(self.pos, self.real_pos)

    @property
    def v_avg(self):
        if self.pos == 'CPT':
            return self.proj / 1.5 - self.average_score
        return self.proj - self.average_score



import os
import random
from draftfast.pickem.pickem_optimize import optimize
from draftfast.csv_parse import uploaders, salary_download
from draftfast import rules
from draftfast.lineup_constraints import LineupConstraints

salary_file_location = os.environ.get('PICKEM')
pid_file_location = os.environ.get('PICKEM_PIDS')


players = salary_download.generate_players_from_csvs(
    game=rules.DRAFT_KINGS,
    salary_file_location=salary_file_location,
    ruleset=rules.DK_NBA_PICKEM_RULE_SET,
)

rosters = []
for p in range(20):
    player = random.choice(players)
    roster = optimize(
        all_players=players,
        constraints=LineupConstraints(locked=[p.name]),
    )
    rosters.append(roster)

uploader = uploaders.DraftKingsNBAPickemUploader(
    pid_file=pid_file_location,
)


uploader.write_rosters(rosters)



import os
from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.csv_parse import salary_download
from draftfast.settings import OptimizerSettings
from draftfast.lineup_constraints import LineupConstraints

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/dk-soccer-salaries.csv'.format(CURRENT_DIR)


def test_soccer_dk():
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_SOCCER_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )
    ntools.assert_not_equal(roster, None)


def test_soccer_dk_no_opp_d():
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
    )
    roster = run(
        rule_set=rules.DK_SOCCER_RULE_SET,
        player_pool=player_pool,
        constraints=LineupConstraints(
            locked=['Maxi Gomez'],
        ),
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=False,
        ),
        verbose=True,
    )
    cel_off_players = [
        p for p in roster.players if p.team == 'CEL'
        and p.pos in ['M', 'F']
    ]
    lgn_d_players = [
        p for p in roster.players if p.team == 'LGN'
        and p.pos in ['D', 'GK']
    ]
    ntools.assert_equal(len(cel_off_players), 2)
    ntools.assert_equal(len(lgn_d_players), 2)

    roster = run(
        rule_set=rules.DK_SOCCER_RULE_SET,
        player_pool=player_pool,
        constraints=LineupConstraints(
            locked=['Maxi Gomez'],
        ),
        optimizer_settings=OptimizerSettings(
            no_offense_against_defense=True,
        ),
        verbose=True,
    )
    cel_off_players = [
        p for p in roster.players if p.team == 'CEL'
        and p.pos in ['M', 'F']
    ]
    lgn_d_players = [
        p for p in roster.players if p.team == 'LGN'
        and p.pos in ['D', 'GK']
    ]
    ntools.assert_equal(len(cel_off_players), 2)
    ntools.assert_equal(len(lgn_d_players), 0)



import os
import csv
from typing import Type
import types
from nose.tools import assert_equal
from draftfast import rules
from draftfast import optimize
from draftfast.csv_parse import uploaders, salary_download
from draftfast.pickem.pickem_optimize import (
    optimize as p_optimize
)

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def test_dk_nba_upload():
    row = _get_first_written_row(
        game=rules.DRAFT_KINGS,
        salary_file_location='{}/data/dk-nba-salaries.csv'.format(CURRENT_DIR),
        rule_set=rules.DK_NBA_RULE_SET,
        pid_file='{}/data/dk-nba-pids.csv'.format(CURRENT_DIR),
        Uploader=uploaders.DraftKingsNBAUploader,
    )
    assert_equal(
        sorted(row),
        sorted([
            '11743190',
            '11743146',
            '11743013',
            '11743142',
            '11743007',
            '11743176',
            '11743369',
            '11743024',
        ])
    )


def test_dk_nfl_upload():
    row = _get_first_written_row(
        game=rules.DRAFT_KINGS,
        salary_file_location='{}/data/dk-nfl-upload-salaries.csv'.format(
            CURRENT_DIR
        ),
        rule_set=rules.DK_NFL_RULE_SET,
        pid_file='{}/data/dk-nfl-upload.csv'.format(CURRENT_DIR),
        Uploader=uploaders.DraftKingsNFLUploader,
    )
    assert_equal(
        row,
        [
            '13651346',
            '13651384',
            '13651396',
            '13651574',
            '13651608',
            '13651652',
            '13651838',
            '13651402',
            '13651964',
        ],
    )


def test_dk_el_upload():
    row = _get_first_written_row(
        game=rules.DRAFT_KINGS,
        salary_file_location='{}/data/dk-euro-league-salaries.csv'.format(
            CURRENT_DIR
        ),
        rule_set=rules.DK_EURO_LEAGUE_RULE_SET,
        pid_file='{}/data/dk-euro-league-pids.csv'.format(CURRENT_DIR),
        Uploader=uploaders.DraftKingsELUploader,
    )
    assert_equal(
        row,
        [
            '11799918',
            '11799942',
            '11799922',
            '11799956',
            '11800052',
            '11799950',
        ],
    )


def test_dk_soccer_upload():
    row = _get_first_written_row(
        game=rules.DRAFT_KINGS,
        salary_file_location='{}/data/dk-soccer-salaries.csv'.format(
            CURRENT_DIR
        ),
        rule_set=rules.DK_SOCCER_RULE_SET,
        pid_file='{}/data/dk-soccer-pids.csv'.format(CURRENT_DIR),
        Uploader=uploaders.DraftKingsSoccerUploader,
    )
    assert_equal(
        row,
        [
            '11801828',
            '11801837',
            '11801757',
            '11801761',
            '11801685',
            '11801733',
            '11801676',
            '11801778',
        ],
    )


def test_fd_nba_upload():
    row = _get_first_written_row(
        game=rules.FAN_DUEL,
        salary_file_location='{}/data/fd-nba-salaries.csv'.format(CURRENT_DIR),
        rule_set=rules.FD_NBA_RULE_SET,
        pid_file='{}/data/fd-nba-pids.csv'.format(CURRENT_DIR),
        Uploader=uploaders.FanDuelNBAUploader,
    )
    assert_equal(
        row,
        [
            '30803-9535:Kyle Lowry',
            '30803-9475:Derrick Rose',
            '30803-9714:DeMar DeRozan',
            '30803-40201:Dennis Schroder',
            '30803-9646:Kevin Durant',
            '30803-12341:Paul George',
            '30803-9957:Serge Ibaka',
            '30803-9874:Kevin Love',
            '30803-12362:DeMarcus Cousins',
        ]
    )


def test_dk_nhl_uploader():
    row = _get_first_written_row(
        game=rules.DRAFT_KINGS,
        salary_file_location='{}/data/dk-nhl-salaries.csv'.format(CURRENT_DIR),
        rule_set=rules.DK_NHL_RULE_SET,
        pid_file='{}/data/dk-nhl-pids.csv'.format(CURRENT_DIR),
        Uploader=uploaders.DraftKingsNHLUploader,
    )
    assert_equal(
        sorted(row),
        sorted([
            '11845288',
            '11845290',
            '11845526',
            '11845550',
            '11845942',
            '11845960',
            '11846094',
            '11846329',
            '11845460',
        ]),
    )


def test_pickem_nba_upload():
    salary_file_location = '{}/data/dk-nba-pickem-salaries.csv'.format(
        CURRENT_DIR
    )
    players = salary_download.generate_players_from_csvs(
        game=rules.DRAFT_KINGS,
        salary_file_location=salary_file_location,
        ruleset=rules.DK_NBA_PICKEM_RULE_SET,
    )
    rosters = [p_optimize(
        all_players=players,
    )]

    pid_file = '{}/data/dk-nba-pickem-pids.csv'.format(CURRENT_DIR)
    upload_file = '{}/data/current-upload.csv'.format(CURRENT_DIR)
    uploader = uploaders.DraftKingsNBAPickemUploader(
        pid_file=pid_file,
        upload_file=upload_file,
    )
    uploader.write_rosters(rosters)

    row = None
    with open(upload_file, 'r') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        for idx, row in enumerate(reader):
            if idx == 0:
                continue
    assert_equal(
        row,
        [
            '11839390',
            '11839397',
            '11839400',
            '11839405',
            '11839420',
            '11839422',
        ]
    )


def test_dk_nfl_showdown_upload():
    row = _get_first_written_row_dk_showdown(
        salary_file='dk-nfl-showdown-salaries.csv',
        pid_file='dk-nfl-showdown-pids.csv',
        ruleset=rules.DK_NFL_SHOWDOWN_RULE_SET,
        Uploader=uploaders.DraftKingsCaptainShowdownUploader,
    )
    assert_equal(
        row,
        [
            '11896024',
            '11895994',
            '11895995',
            '11895996',
            '11895997',
            '11896018',
        ]
    )


def test_dk_nba_showdown_upload():
    row = _get_first_written_row_dk_showdown(
        salary_file='dk-nba-showdown-salaries.csv',
        pid_file='dk-nba-showdown-pids.csv',
        ruleset=rules.DK_NBA_SHOWDOWN_RULE_SET,
        Uploader=uploaders.DraftKingsCaptainShowdownUploader,
    )
    assert_equal(
        row,
        [
            '11915867',
            '11915844',
            '11915845',
            '11915846',
            '11915852',
            '11915853',
        ],
    )


def test_dk_mlb_showdown_upload():
    row = _get_first_written_row_dk_showdown(
        salary_file='dk-mlb-showdown-salaries.csv',
        pid_file='dk-mlb-showdown-pids.csv',
        ruleset=rules.DK_MLB_SHOWDOWN_RULE_SET,
        Uploader=uploaders.DraftKingsCaptainShowdownUploader,
    )
    assert_equal(
        row,
        [
            '12895602',
            '12895494',
            '12895495',
            '12895496',
            '12895510',
            '12895600',
        ],
    )


def test_dk_xfl_classic_upload():
    def proj(players):
        for idx, p in enumerate(players):
            p.proj = idx

    row = _get_first_written_row_dk_showdown(
        salary_file='dk-xfl-salaries.csv',
        pid_file='dk-xfl-pids.csv',
        ruleset=rules.DK_XFL_CLASSIC_RULE_SET,
        Uploader=uploaders.DraftKingsXFLUploader,
        players_side_effect=proj
    )
    assert_equal(
        row,
        [
            '14224920',
            '14225002',
            '14225226',
            '14225228',
            '14225230',
            '14225232',
            '14225241'
        ]
    )


def _get_first_written_row(
        game: str,
        salary_file_location: str,
        rule_set: rules.RuleSet,
        pid_file: str,
        Uploader: Type[uploaders.CSVUploader],
) -> list:
    players = salary_download.generate_players_from_csvs(
        game=game,
        salary_file_location=salary_file_location,
        ruleset=rule_set,
    )
    roster = optimize.run(
        rule_set=rule_set,
        player_pool=players,
        verbose=True,
    )
    upload_file = '{}/data/current-upload.csv'.format(CURRENT_DIR)
    uploader = Uploader(
        pid_file=pid_file,
        upload_file=upload_file,
    )
    uploader.write_rosters([roster])

    row = None
    with open(upload_file, 'r') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        for idx, row in enumerate(reader):
            if idx == 0:
                continue

    return row


def _noop(players):
    pass


def _get_first_written_row_dk_showdown(
    salary_file: str,
    pid_file: str,
    ruleset: rules.RuleSet,
    Uploader: Type[uploaders.CSVUploader],
    players_side_effect: types.FunctionType = _noop
) -> list:
    salary_file_location = '{}/data/{}'.format(
        CURRENT_DIR,
        salary_file,
    )
    players = salary_download.generate_players_from_csvs(
        game=rules.DRAFT_KINGS,
        salary_file_location=salary_file_location,
        ruleset=ruleset,
    )
    players_side_effect(players)
    rosters = [optimize.run(
        player_pool=players,
        rule_set=ruleset,
        verbose=True,
    )]

    pid_file_location = '{}/data/{}'.format(CURRENT_DIR, pid_file)
    upload_file = '{}/data/current-upload.csv'.format(CURRENT_DIR)
    uploader = Uploader(
        pid_file=pid_file_location,
        upload_file=upload_file,
    )
    uploader.write_rosters(rosters)

    row = None
    with open(upload_file, 'r') as csvfile:
        reader = csv.reader(csvfile, delimiter=',')
        for idx, row in enumerate(reader):
            if idx == 0:
                continue

    return row



import csv
from draftfast.orm import Player
from draftfast.pickem.pickem_orm import TieredPlayer
from draftfast.showdown.orm import ShowdownPlayer
from draftfast.rules import DRAFT_KINGS, FAN_DUEL

GAME_KEY_MAP = {
    DRAFT_KINGS: {
        'name': 'Name',
        'salary': 'Salary',
        'team': 'teamAbbrev',
        'team_alt': 'TeamAbbrev',
        'game': 'GameInfo',
        'game_alt': 'Game Info',
        'avg': 'AvgPointsPerGame',
    },
    FAN_DUEL: {
        'name': 'Nickname',
        'team': 'Team',
        'team_alt': None,
        'game': 'Game',
        'game_alt': None,
        'avg': 'FPPG',
    }
}


# TODO - return players given two CSV locations and RuleSet
# The two CSVs here should be 1) the DK / FD salary sheet
# and 2) player projections
def generate_players_from_csvs(
    salary_file_location: str,
    game: str,
    projection_file_location='',
    verbose=False,
    encoding='utf-8',
    errors='replace',
    ruleset=None,
) -> list:
    players = []
    projections = None
    if projection_file_location:
        projections = _generate_projection_dict(
            projection_file_location,
            encoding,
            errors,
        )

    with open(salary_file_location, 'r',
              encoding=encoding, errors=errors) as csv_file:
        csv_data = csv.DictReader(csv_file)
        pos_key = 'Position'
        is_nhl = ruleset and ruleset.league == 'NHL'
        is_showdown = ruleset and ruleset.game_type == 'showdown'
        if is_nhl or is_showdown:
            pos_key = 'Roster Position'

        for row in csv_data:
            if ruleset and ruleset.game_type == 'pickem':
                player = TieredPlayer(
                    cost=0,  # salary not applicable in pickem
                    name=row['Name'],
                    pos=row['Position'],
                    team=row['TeamAbbrev'],
                    matchup=row['Game Info'],
                    average_score=float(row['AvgPointsPerGame']),
                    tier=row['Roster Position'],
                )
                _set_projections(
                    projections,
                    player,
                    verbose,
                )
                players.append(player)
            elif is_showdown:
                pos = row[pos_key]
                player = generate_player(
                    pos=row['Position'],
                    row=row,
                    game=game,
                )
                _set_projections(
                    projections,
                    player,
                    verbose,
                )
                captain = pos == 'CPT'
                players.append(
                    ShowdownPlayer(
                        player,
                        captain=captain,
                    )
                )
            else:
                for pos in row[pos_key].split('/'):
                    if is_nhl and pos == 'UTIL':
                        continue
                    player = generate_player(
                        pos=pos,
                        row=row,
                        game=game,
                    )
                    _set_projections(
                        projections,
                        player,
                        verbose,
                    )

                    players.append(player)

    return players


# TODO - extract
def _create_classic_player():
    pass


def _create_tiered_player():
    pass


def _create_showdown_player():
    pass


def generate_player(pos, row, game):
    '''
    Parses CSV row for DraftKings or FanDuel
    and returns a player. Note that DraftKings
    has different CSV formats for different
    sports.
    '''
    avg_key = GAME_KEY_MAP[game]['avg']
    name_key = GAME_KEY_MAP[game]['name']
    team_key = GAME_KEY_MAP[game]['team']
    team_alt_key = GAME_KEY_MAP[game]['team_alt']
    game_key = GAME_KEY_MAP[game]['game']
    game_alt_key = GAME_KEY_MAP[game]['game_alt']

    avg = float(row.get(avg_key, 0))

    player = Player(
        pos,
        row[name_key].strip().rstrip(),
        row['Salary'],
        possible_positions=row['Position'],
        multi_position='/' in row['Position'],
        team=row.get(team_key) or row.get(team_alt_key),
        matchup=row.get(game_key) or row.get(game_alt_key),
        average_score=avg,
        kv_store=row,
    )

    return player


def _generate_projection_dict(projection_file_location: str,
                              encoding: str,
                              errors: str) -> dict:
    projections = {}
    with open(projection_file_location, 'r',
              encoding=encoding, errors=errors) as csv_file:
        csv_data = csv.DictReader(csv_file)
        for row in csv_data:
            name = row.get('playername').strip().rstrip()
            projections[name] = float(row.get('points'))

    return projections


def _set_projections(projections, player, verbose):
    if projections:
        proj = projections.get(player.name)

        # try to find projection for e.g. 'Andrew Luck IND'
        if proj is None and player.team is not None:
            alt_name = '{} {}'.format(player.name, player.team.upper())
            proj = projections.get(alt_name)

        if proj is None:
            if verbose:
                print('No projection for {}'.format(player.name))
            player.proj = 0
        else:
            player.proj = proj
    else:
        player.proj = player.average_score



import os
from nose import tools as ntools
from draftfast.csv_parse import salary_download
from draftfast.rules import DRAFT_KINGS

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salaries = '{}/data/nba-test-salaries.csv'.format(CURRENT_DIR)
projections = '{}/data/nba-test-projections.csv'.format(CURRENT_DIR)


def test_dk_nba_parse():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salaries,
        game=DRAFT_KINGS,
    )
    ntools.assert_equals(len(players), 221)


def test_dk_nba_use_avg():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salaries,
        game=DRAFT_KINGS,
    )
    ntools.assert_equals(players[0].proj, 60.462)


def test_dk_nba_use_proj():
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salaries,
        projection_file_location=projections,
        game=DRAFT_KINGS,
    )
    ntools.assert_equals(players[0].proj, 62.29)



from nose import tools as ntools
from draftfast.pickem import pickem_orm
from draftfast.pickem.pickem_optimize import optimize
from draftfast.settings import PlayerPoolSettings
from draftfast.lineup_constraints import LineupConstraints


def _generate_player(name, proj, tier, **kwargs):
    return pickem_orm.TieredPlayer(
        cost=0,
        name=name,
        proj=proj,
        tier=tier,
        pos=kwargs.get('pos'),
        team=kwargs.get('team'),
        matchup=kwargs.get('matchup'),
        average_score=kwargs.get('average_score', 0),
    )


_BOS = 'BOS'
_GS = 'GS'


def _generate_test_player_data():
    return [
        _generate_player('A', 50, pickem_orm.T1, team=_BOS),
        _generate_player('B', 49, pickem_orm.T1, team=_GS),
        _generate_player('C', 48, pickem_orm.T2, team=_BOS),
        _generate_player('D', 47, pickem_orm.T2, team=_GS),
        _generate_player('E', 46, pickem_orm.T3, team=_BOS),
        _generate_player('F', 45, pickem_orm.T3, team=_GS),
        _generate_player('G', 44, pickem_orm.T4, team=_BOS),
        _generate_player('H', 43, pickem_orm.T4, team=_GS),
        _generate_player('I', 42, pickem_orm.T5, team=_BOS),
        _generate_player('J', 41, pickem_orm.T5, team=_GS),
        _generate_player('K', 40, pickem_orm.T6, team=_BOS),
        _generate_player('L', 39, pickem_orm.T6, team=_GS),
    ]


def test_default_lineup():
    players = _generate_test_player_data()
    optimized = optimize(players)
    ntools.assert_equal(
        optimized.total,
        50 + 48 + 46 + 44 + 42 + 40
    )


def test_banned_players():
    players = _generate_test_player_data()

    optimized = optimize(
        players,
        player_settings=PlayerPoolSettings(),
        constraints=LineupConstraints(banned=['A', 'C'])
    )
    ntools.assert_equal(
        optimized.total,
        49 + 47 + 46 + 44 + 42 + 40
    )


def test_locked_players():
    players = _generate_test_player_data()

    optimized = optimize(
        players,
        constraints=LineupConstraints(locked=['B', 'D'])
    )
    ntools.assert_equal(
        optimized.total,
        49 + 47 + 46 + 44 + 42 + 40
    )
#
#
# def test_banned_teams():
#     players = _generate_test_player_data()
#     test_args = args_dict.copy()
#     test_args['banned_teams'] = [_BOS]
#     optimized = optimize(players, cmd_args=Namespace(**test_args))
#     ntools.assert_equal(
#         optimized.total,
#         49 + 47 + 45 + 43 + 41 + 39
#     )
#
#
# def test_locked_teams():
#     players = _generate_test_player_data()
#     test_args = args_dict.copy()
#     test_args['locked_teams'] = [_GS]
#     optimized = optimize(players, cmd_args=Namespace(**test_args))
#     ntools.assert_equal(
#         optimized.total,
#         49 + 47 + 45 + 43 + 41 + 39
#     )



import os
from nose import tools as ntools
from draftfast.optimize import run_multi
from draftfast import rules
from draftfast.csv_parse import salary_download

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/dk-nfl-salaries.csv'.format(CURRENT_DIR)
fd_nfl_salary_file = '{}/data/fd-nfl-salaries.csv'.format(CURRENT_DIR)
projection_file = '{}/data/dk-nfl-projections.csv'.format(CURRENT_DIR)


def test_deterministic_exposure_limits():
    iterations = 2
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    rosters, exposure_diffs = run_multi(
        iterations=2,
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
        exposure_bounds=[
            {'name': 'Andrew Luck', 'min': 0.5, 'max': 0.7},
            {'name': 'Alshon Jeffery', 'min': 1, 'max': 1},
        ],
    )
    ntools.assert_equal(len(rosters), iterations)
    ntools.assert_equal(len(exposure_diffs), 0)

    players = [p.name for p in rosters[0].players]
    ntools.assert_true('Andrew Luck' in players)
    ntools.assert_true('Alshon Jeffery' in players)

    players = [p.name for p in rosters[1].players]
    ntools.assert_true('Andrew Luck' not in players)
    ntools.assert_true('Alshon Jeffery' in players)


def test_random_exposure_limits():
    iterations = 10
    players = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        projection_file_location=projection_file,
        game=rules.DRAFT_KINGS,
    )
    rosters, exposure_diffs = run_multi(
        iterations=iterations,
        exposure_random_seed=42,
        rule_set=rules.DK_NFL_RULE_SET,
        player_pool=players,
    )
    ntools.assert_equal(len(rosters), iterations)
    ntools.assert_equal(len(exposure_diffs), 0)



import locale
from terminaltables import AsciiTable
from functools import total_ordering
import re

try:
    locale.setlocale(locale.LC_ALL, 'en_US')
except Exception:
    pass


def cs(n):
    return locale.format('%d', n, grouping=True)


class Roster:
    def __init__(self):
        self.players = []
        self.cached_id = None

    def __repr__(self):
        table_data = []
        headers = [
            'Position',
            'Player',
            'Team',
            'Matchup',
            'Salary',
            'Projection',
            'vs. Avg.',
            'Locked'
        ]
        table_data.append(headers)
        for p in self.sorted_players():
            table_data.append(p.to_table_row())

        table = AsciiTable(table_data)
        table.justify_columns[4] = 'right'
        table.justify_columns[5] = 'right'
        table.justify_columns[6] = 'right'

        aggregate_info = '\n\nProjected Score: {:0.2f} \t Cost: ${}'.format(
            self.projected(),
            cs(self.spent()))

        return table.table + aggregate_info

    @property
    def identifier(self):
        if self.cached_id:
            return self.cached_id
        self.cached_id = ' '.join(sorted([
            x.roster_id for x in self.sorted_players()
        ]))

        return self.cached_id

    def __eq__(self, other):
        if not other:
            return False

        return self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier)

    def __contains__(self, player):
        if isinstance(player, str):
            for p in self.players:
                if p.name == player or p.short_name == player:
                    return True
        elif isinstance(player, Player):
            return player in self.players
        else:
            raise NotImplementedError

        return False

    def add_player(self, player):
        self.players.append(player)

    def spent(self):
        return sum([x.cost for x in self.players])

    def projected(self):
        return sum([x.proj for x in self.players])

    def position_order(self, player):
        # raises exception in case someone tries to instantiate base class
        position_order = getattr(self, 'POSITION_ORDER')

        # default sort order is low->high, so use the negative of cost
        return position_order[player.pos], -player.cost

    def sorted_players(self):
        return sorted(
            self.players,
            key=lambda p: self.position_order(p)
        )


'''
POSITION_ORDER is based on the order
required by DraftKings' CSV download
'''


class ShowdownRoster(Roster):
    POSITION_ORDER = {
        'CPT': 0,
        'FLEX': 1,
    }

    @property
    def identifier(self):
        """
        In Showdown, only two positions exist
        and a change in position means a change
        in points, so unique is on position.
        """
        if self.cached_id:
            return self.cached_id
        self.cached_id = ' '.join(sorted([
            x.solver_id for x in self.sorted_players()
        ]))

        return self.cached_id


class NFLRoster(Roster):
    POSITION_ORDER = {
        'QB': 0,
        'RB': 1,
        'WR': 2,
        'TE': 3,
        'DST': 4,
        'D': 5,
    }


class TenRoster(Roster):
    POSITION_ORDER = {
        'P': 1,
    }


class MLBRoster(Roster):
    POSITION_ORDER = {
        'P': 0,
        'RP': 0,
        'SP': 0,
        'C': 1,
        '1B': 2,
        '2B': 3,
        '3B': 4,
        'SS': 5,
        'OF': 6,
    }


class NBARoster(Roster):
    POSITION_ORDER = {
        'PG': 0,
        'SG': 1,
        'SF': 2,
        'PF': 3,
        'C': 4
    }


class WNBARoster(Roster):
    POSITION_ORDER = {
        'G': 0,
        'F': 1,
        'SG': 2,
        'SF': 3,
        'PF': 4,
    }


class NASCARRoster(Roster):
    POSITION_ORDER = {
        'D': 0,
    }


class PGARoster(Roster):
    POSITION_ORDER = {
        'G': 0,
    }


class SoccerRoster(Roster):
    POSITION_ORDER = {
        'F': 0,
        'M': 1,
        'D': 2,
        'GK': 3,
    }


class ELRoster(Roster):
    POSITION_ORDER = {
        'G': 0,
        'F': 1,
    }


class NHLRoster(Roster):
    POSITION_ORDER = {
        'C': 0,
        'W': 1,
        'D': 2,
        'G': 3,
    }


class RosterSelect:
    @staticmethod
    def roster_gen(league):
        roster_dict = {
            'NBA': NBARoster(),
            'NBA_SHOWDOWN': ShowdownRoster(),
            'WNBA': WNBARoster(),
            'NFL': NFLRoster(),
            'NFL_SHOWDOWN': ShowdownRoster(),
            'NFL_MVP': ShowdownRoster(),
            'MLB': MLBRoster(),
            'PGA': PGARoster(),
            'NASCAR': NASCARRoster(),
            'SOCCER': SoccerRoster(),
            'EL': ELRoster(),
            'NHL': NHLRoster(),
            'NHL_SHOWDOWN': ShowdownRoster(),
            'MLB_SHOWDOWN': ShowdownRoster(),
            # XFL uses the same positions as NFL
            'XFL': NFLRoster(),
            'TEN': TenRoster(),
            'CSGO_SHOWDOWN': ShowdownRoster(),
        }
        return roster_dict[league]


@total_ordering
class Player(object):
    def __init__(
        self,
        pos,
        name,
        cost,
        proj=0,
        projected_ownership_pct=0,
        lineup_count=0,
        average_score=0,
        matchup=None,
        team=None,
        marked=None,
        possible_positions=None,
        lock=False,
        position_lock=False,
        ban=False,
        multi_position=False,
        kv_store={},
    ):
        self.pos = pos
        self.name = name
        self.cost = float(cost)
        self.team = team.upper() if team else team
        self.matchup = matchup
        self.proj = proj
        self.average_score = average_score
        self.projected_ownership_pct = projected_ownership_pct
        self.lineup_count = lineup_count
        self.marked = marked
        self.lock = lock
        self.position_lock = position_lock
        self.ban = ban
        self.position_ban = False
        self.multi_position = multi_position
        self.possible_positions = possible_positions
        self.kv_store = kv_store

    def get_player_id(self, player_map):
        return player_map[self.name + ' ' + self.possible_positions]

    def to_table_row(self):
        return [
            self.formatted_position,
            self.name,
            self.team,
            self.matchup,
            cs(self.cost),
            self.proj,
            self.__format_v_avg(),
            'LOCK' if self.lock else ''
        ]

    def to_exposure_table_row(self, n, s_min, s_max):
        return [
            self.formatted_position,
            self.name,
            self.team,
            self.matchup,
            cs(self.cost),
            self.proj,
            n,
            s_min,
            s_max
        ]

    def is_opposing_team_in_match_up(self, team):
        return (team.upper() != self.team.upper()) and \
               (team.upper() in self.matchup.upper())

    def __repr__(self):
        v_avg = self.__format_v_avg()
        player_dict = dict(
            pos=self.formatted_position,
            name=self.name,
            team=self.team,
            match=self.matchup,
            cost=cs(self.cost),
            proj=self.proj,
            v_avg=v_avg,
            lock='LOCK' if self.lock else ''
        )

        return "[{pos: <2}] {name: <20} {team} {match} " \
               "(${cost}, {proj} ({v_avg})), {lock}".format(
                   **player_dict)

    def __eq__(self, player):
        return self.pos == player.pos and \
            self.name == player.name and \
            self.cost == player.cost and \
            self.team == player.team

    def __hash__(self):
        return hash((self.pos, self.name, self.cost, self.team))

    def __lt__(self, player):
        if self.cost == player.cost:
            return self.name < player.name
        return self.cost < player.cost

    @property
    def value(self):
        return round(self.proj / (self.cost / 1000), 2)

    @property
    def solver_id(self):
        """
        Used by optimizer to accommodate same player in different positions
        """
        return '{} {} {}'.format(self.name, self.pos, self.team)

    @property
    def roster_id(self):
        """
        Used for roster equality. From the perspective of someone
        entering lineups,a lineup with the same players scores
        the same points irrespective of positions
        """
        return f'{self.name} {self.team}'

    @property
    def formatted_position(self):
        if self.multi_position:
            return '{} ({})'.format(self.possible_positions, self.pos)
        return self.pos

    @property
    def v_avg(self):
        return self.proj - self.average_score

    @property
    def is_home(self):
        match_up_teams = self.matchup.split(' ')[0]
        return self.team == match_up_teams.split('@')[-1]

    @property
    def nba_general_position(self):
        if self.pos == 'SG' or self.pos == 'PG' or self.pos == 'G':
            return 'G'
        elif self.pos == 'SF' or self.pos == 'PF' or self.pos == 'F':
            return 'F'
        return 'C'

    @property
    def mlb_general_position(self):
        if self.pos in {'SP', 'RP'}:
            return 'P'
        return self.pos

    @property
    def short_name(self):
        s = self.name.split()

        # DST
        if len(s) == 1:
            return self.name

        # like "AJ McCarron"
        if re.match(r'^[A-Z]{2}$', s[0]):
            return s

        return '{}. {}'.format(s[0][0], s[1])

    def __set_from_data_cache(self, player_data):
        if player_data is None:
            return
        for k, v in list(player_data.items()):
            setattr(self, k, v)

    def __format_v_avg(self):
        if self.v_avg > 0:
            return '\x1b[0;32;40m{:0.2f}\x1b[0m'.format(self.v_avg)
        return '\x1b[0;31;40m{:0.2f}\x1b[0m'.format(self.v_avg)


class Game:
    def __init__(self, team, opp):
        self.team = team
        self.opponent = opp

    def __repr__(self):
        return '{} @ {}'.format(self.team, self.opponent)

    def team_in_game(self, team):
        return team == self.team or team == self.opponent

    def get_teams(self):
        return self.team, self.opponent



from random import uniform as runiform
from typing import List
from draftfast.orm import Player
from draftfast.settings import PlayerPoolSettings


def filter_pool(pool: list,
                player_settings: PlayerPoolSettings) -> List[Player]:
    if player_settings.randomize:
        for player in pool:
            factor = 1 + runiform(
                -player_settings.randomize,
                player_settings.randomize
            )
            player.proj = player.proj * factor

    return list(filter(
        add_filters(player_settings),
        pool,
    ))


def add_filters(settings: PlayerPoolSettings):
    def filter_fn(player: Player):
        kwargs = {'player': player, 'settings': settings}
        return _is_above_min_cost(**kwargs) and \
            _is_below_max_cost(**kwargs) and \
            _is_above_min_proj(**kwargs) and \
            _is_below_max_proj(**kwargs) and \
            _is_above_min_avg(**kwargs) and \
            _is_below_max_avg(**kwargs)

    return filter_fn


def add_pickem_contraints(settings: PlayerPoolSettings):
    def filter_fn(player: Player):
        # TODO - add team banning
        kwargs = {'player': player, 'settings': settings}
        return (
            _is_above_min_proj(**kwargs) and
            # (not _is_banned_team(**kwargs)) and
            # _is_locked_team(**kwargs) and
            # _is_within_avg(**kwargs) and
            _is_above_min_avg(**kwargs)
        )

    return filter_fn


def lock_override(fn):
    def override_fn(**kwargs):
        if kwargs['player'].lock:
            return True
        return fn(**kwargs)

    return override_fn


@lock_override
def _is_above_min_cost(player, settings):
    if settings.min_salary is None:
        return True
    return player.cost >= settings.min_salary


@lock_override
def _is_below_max_cost(player, settings):
    if settings.max_salary is None:
        return True
    return player.cost <= settings.max_salary


@lock_override
def _is_above_min_proj(player, settings):
    if settings.min_proj is None:
        return True
    return player.proj >= settings.min_proj


@lock_override
def _is_below_max_proj(player, settings):
    if settings.max_proj is None:
        return True
    return player.proj <= settings.max_proj


@lock_override
def _is_above_min_avg(player, settings):
    if settings.min_avg is None:
        return True
    return player.average_score >= settings.min_avg


@lock_override
def _is_below_max_avg(player, settings):
    if settings.max_avg is None:
        return True
    return player.average_score <= settings.max_avg



from draftfast import (
    optimize,
    rules,
    orm,
    csv_parse,
    exposure,
    pickem,
    showdown,
)

assert optimize
assert rules
assert orm
assert csv_parse
assert exposure
assert pickem
assert showdown



import math
import csv
import random
from collections import OrderedDict
from terminaltables import AsciiTable
import numpy as np

# TODO encapsulate this into an object


def parse_exposure_file(file_location):
    """
    :param file: File location
    :return: Dictionary of exposures
    { <name>: { min: <min>, max: <max> } }
    """
    exposures = []
    with open(file_location, 'r') as filename:
        reader = csv.DictReader(filename)
        for row in reader:
            if 'name' not in row or \
               'min' not in row or \
               'max' not in row:
                raise Exception('''
                    You must provide a min, max and name
                    for each row - got {}.
                    '''.format(row)
                )
            exposures.append({
                'name': row['name'],
                'min': float(row['min']),
                'max': float(row['max']),
            })

    return exposures


def get_exposure_args(existing_rosters, exposure_bounds, n, use_random,
                      random_seed) -> dict:
    exposures = {}
    for r in existing_rosters:
        for p in r.players:
            exposures[p.name] = exposures.get(p.name, 0) + 1

    if use_random:
        return get_exposure_args_random(exposures, exposure_bounds, n,
                                        random_seed)

    return get_exposure_args_deterministic(exposures, existing_rosters,
                                           exposure_bounds)


def get_exposure_args_deterministic(exposures, existing_rosters,
                                    exposure_bounds) -> dict:
    banned = []
    locked = []

    for bound in exposure_bounds:
        name = bound['name']

        total = float(len(existing_rosters) + 1)
        min_lines = bound['min'] * total
        max_lines = math.floor(bound['max'] * total)
        lineups = exposures.get(name, 0)

        if lineups < min_lines:
            # TODO - downsize locked so solution is not impossible
            locked.append(name)
        elif lineups >= max_lines:
            banned.append(name)

    return {
        'banned': banned,
        'locked': locked,
    }


def get_exposure_args_random(exposures, exposure_bounds, n,
                             random_seed) -> dict:
    locked = []

    for bound in exposure_bounds:
        name = bound['name']

        # TODO: maybe exclude players who have met max exposure?
        # randomly lock in players based on the desired exposure
        # TODO - downsize locked so solution is not impossible
        r = random.random()
        if r <= bound['max']:
            locked.append(name)

    return {
        'banned': [],
        'locked': locked,
    }


# TODO split this up to return total exposures, exposure_diffs
def check_exposure(rosters, bounds):
    if not bounds:
        return {}

    exposures = {}
    for r in rosters:
        for p in r.players:
            exposures[p.name] = exposures.get(p.name, 0) + 1

    exposure_diffs = {}

    for bound in bounds:
        name = bound['name']
        exposure = exposures.get(name, 0)

        if exposure > len(rosters) * bound['max']:
            exposure_diffs[name] = exposure - len(rosters) * bound['max']
        elif exposure < len(rosters) * bound['min']:
            exposure_diffs[name] = exposure - len(rosters) * bound['min']

    return exposure_diffs


def get_exposure_table(rosters, bounds):
    exposures = {}
    players = {}
    for r in rosters:
        for p in r.players:
            exposures[p.name] = exposures.get(p.name, 0) + 1
            players[p.name] = p

    exposures = OrderedDict(sorted(exposures.items(),
                                   key=lambda t: t[1],
                                   reverse=True))

    table_data = []
    headers = [
        'Position',
        'Player',
        'Team',
        'Matchup',
        'Salary',
        'Projection',
        '# Lineups',
        'Min',
        'Max'
    ]
    table_data.append(headers)

    for name, num in exposures.items():
        s_min = ''
        s_max = ''

        # TODO format min/max as a single string
        if bounds:
            for bound in bounds:
                if bound['name'] == name:
                    s_min = len(rosters) * bound['min']
                    s_max = len(rosters) * bound['max']
                    if num > s_max:
                        s_max = '\x1b[0;31;40m{:0.2f}\x1b[0m'.format(s_max)
                    elif num < s_min:
                        s_min = '\x1b[0;31;40m{:0.2f}\x1b[0m'.format(s_min)

                    continue

        table_data.append(players[name].to_exposure_table_row(num,
                                                              s_min,
                                                              s_max))

    table = AsciiTable(table_data)
    table.justify_columns[4] = 'right'
    table.justify_columns[5] = 'right'
    table.justify_columns[6] = 'right'
    table.justify_columns[7] = 'right'
    table.justify_columns[8] = 'right'

    return 'Roster Exposure:\n' + table.table


def get_exposure_matrix(rosters, exclude=[]):
    players = set()
    for r in rosters:
        for p in r.players:
            if p in exclude:
                continue
            players.add(p)

    sorted_names = sorted([p.short_name for p in players])
    player_matrix = np.zeros((len(players), len(players)), dtype=int)

    for r in rosters:
        for i, p1 in enumerate(sorted_names):
            for j, p2 in enumerate(sorted_names):
                if p1 in r and p2 in r:
                    player_matrix[i, j] += 1

    rows = [[''] + sorted_names]

    for i, p in enumerate(sorted_names):
        rows.append([p] + list(player_matrix[i, :]))

    table = AsciiTable(rows)
    table.inner_row_border = True
    table.justify_columns = {i + 1: 'center' for i in range(len(sorted_names))}

    return table.table



from draftfast.lineup_constraints import (LineupConstraints,
                                          ConstraintConflictException,
                                          ConstraintException)
from nose import tools as ntools


def test_constraint_string_args():
    lcs = LineupConstraints()
    lcs.ban('Sam Bradford')
    lcs.lock('Will Fuller')
    ntools.assert_equal(len(lcs), 2)


def test_constraint_contains():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['A', 'B'], 1)
    lcs.ban(['C'])
    lcs.add_group_constraint(['E', 'F', 'G'], (1, 3))
    lcs.lock(['H'])

    for c in ['A', 'B', 'C', 'E', 'F', 'G', 'H']:
        ntools.assert_equal(c in lcs, True)


def test_constraint_set_eq():
    lcs1 = LineupConstraints()
    lcs1.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 1)
    lcs1.ban(['Packers'])
    lcs1.add_group_constraint(['Eli Manning', 'Russell Wilson', 'Doug Martin'],
                              (1, 3))
    lcs1.lock(['Will Fuller'])

    lcs2 = LineupConstraints()
    lcs2.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 1)
    lcs2.ban(['Packers'])
    lcs2.add_group_constraint(['Eli Manning', 'Russell Wilson', 'Doug Martin'],
                              (1, 3))
    lcs2.lock(['Will Fuller'])

    ntools.assert_equal(lcs1, lcs2)


def test_build_constraint_set():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 1)
    lcs.ban(['Packers'])
    lcs.add_group_constraint(['Eli Manning', 'Russell Wilson', 'Doug Martin'],
                             (1, 3))
    lcs.lock(['Will Fuller'])

    ntools.assert_equal(len(lcs), 4)


@ntools.raises(ConstraintConflictException)
def test_dup_group_rule():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 1)
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 1)


@ntools.raises(ConstraintConflictException)
def test_dup_group_rule2():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Eli Manning', 'Russell Wilson', 'Doug Martin'],
                             (1, 2))
    lcs.add_group_constraint(['Eli Manning', 'Russell Wilson', 'Doug Martin'],
                             (1, 2))


@ntools.raises(ConstraintException)
def test_bad_group_shadow_lock_bound():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 2)


@ntools.raises(ConstraintException)
def test_bad_group_shadow_lock_hi_lo_bound():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (2, 2))


@ntools.raises(ConstraintException)
def test_bad_group_shadow_ban_bound():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 0)


@ntools.raises(ConstraintException)
def test_bad_group_shadow_ban_hi_lo_bound():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (0, 0))


@ntools.raises(ConstraintException)
def test_bad_group_duplicate_bounds():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (1, 1))


@ntools.raises(ConstraintException)
def test_bad_group_negative_min():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (-1, 1))


@ntools.raises(ConstraintException)
def test_bad_group_zero_min():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (0, 1))


@ntools.raises(ConstraintException)
def test_bad_group_max():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], 3)


@ntools.raises(ConstraintException)
def test_bad_group_max_set():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (1, 3))


@ntools.raises(ConstraintException)
def test_bad_group_dup_player():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Amari Cooper', 'Amari Cooper'], 1)


@ntools.raises(ConstraintException)
def test_bad_group_bounds_type():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], '1')


@ntools.raises(ConstraintException)
def test_bad_group_too_many_bounds():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware', 'Amari Cooper'], (1, 1, 2))


@ntools.raises(ConstraintException)
def test_single_player_group():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Spencer Ware'], 1)


@ntools.raises(ConstraintException)
def test_empty_group():
    lcs = LineupConstraints()
    lcs.add_group_constraint([], 1)


@ntools.raises(ConstraintException)
def test_empty_lock():
    lcs = LineupConstraints()
    lcs.lock([])


@ntools.raises(ConstraintException)
def test_empty_ban():
    lcs = LineupConstraints()
    lcs.ban([])


@ntools.raises(ConstraintConflictException)
def test_ban_lock_conflict():
    lcs = LineupConstraints()
    lcs.lock(['Will Fuller'])
    lcs.ban(['Will Fuller'])


@ntools.raises(ConstraintConflictException)
def test_lock_ban_conflict():
    lcs = LineupConstraints()
    lcs.ban(['Will Fuller'])
    lcs.lock(['Will Fuller'])


@ntools.raises(ConstraintConflictException)
def test_lock_group_conflict():
    lcs = LineupConstraints()
    lcs.lock(['Eli Manning'])
    lcs.add_group_constraint(['Eli Manning', 'Doug Martin'], (1, 2))


@ntools.raises(ConstraintConflictException)
def test_group_lock_conflict():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Eli Manning', 'Doug Martin'], (1, 2))
    lcs.lock(['Eli Manning'])


@ntools.raises(ConstraintConflictException)
def test_ban_group_conflict():
    lcs = LineupConstraints()
    lcs.ban(['Eli Manning'])
    lcs.add_group_constraint(['Eli Manning', 'Doug Martin'], (1, 2))


@ntools.raises(ConstraintConflictException)
def test_group_ban_conflict():
    lcs = LineupConstraints()
    lcs.add_group_constraint(['Eli Manning', 'Doug Martin'], (1, 2))
    lcs.ban(['Eli Manning'])



class MissingPlayersException(Exception):
    pass


class InvalidNFLTeamException(Exception):
    pass


class InvalidCSVUploadFileException(Exception):
    pass


class InvalidBoundsException(Exception):
    pass


class PlayerBanAndLockException(Exception):
    pass


MISSING_ERROR = """
Got {} projections out of {} total players.

You are allowing {} players to be missing from your
projections compared to the total players DraftKings
will allow you to play. You can change this allowance
via the mp flag.
"""

CSV_ERROR = """
You must provide the following fields
in your projection CSV: {}
"""



import os
from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.csv_parse import salary_download
from draftfast.orm import Player
from draftfast.settings import CustomRule, OptimizerSettings

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
salary_file = '{}/data/dk-mlb-salaries.csv'.format(CURRENT_DIR)


def test_mlb_dk():
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
        ruleset=rules.DK_MLB_RULE_SET,
    )
    roster = run(
        rule_set=rules.DK_MLB_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )

    # Test general position limits
    ntools.assert_not_equal(roster, None)
    ntools.assert_true('RP' in [x.pos for x in roster.players])


def test_five_batters_max():
    player_pool = [
        Player(pos='P', name='A', cost=5000, team='C'),
        Player(pos='P', name='B', cost=5000, team='B'),

        Player(pos='1B', name='C', cost=5000, team='C'),
        Player(pos='OF', name='H', cost=5000, team='C'),
        Player(pos='OF', name='I', cost=5000, team='C'),
        Player(pos='C', name='F', cost=5000, team='C'),
        Player(pos='2B', name='D', cost=5000, team='C'),
        Player(pos='2B', name='E', cost=5000, team='C'),
        Player(pos='3B', name='E', cost=5000, team='C'),

        Player(pos='SS', name='G', cost=5000, team='Q'),
        Player(pos='OF', name='J', cost=5000, team='G'),
    ]

    roster = run(
        rule_set=rules.DK_MLB_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )
    assert roster is None

    player_pool.append(Player(pos='3B', name='EA', cost=5000, team='A'))
    roster = run(
        rule_set=rules.DK_MLB_RULE_SET,
        player_pool=player_pool,
        verbose=True,
    )
    c_in_roster = [
        x for x in roster.players
        if x.team == 'C'
        and x.pos != 'P'
    ]

    assert len(c_in_roster) < 6


def test_custom_rules():
    # Minimum stack size
    custom_rules = []
    player_pool = salary_download.generate_players_from_csvs(
        salary_file_location=salary_file,
        game=rules.DRAFT_KINGS,
        ruleset=rules.DK_MLB_RULE_SET,
    )
    for p in player_pool:
        if p.team == 'ATL' and p.pos == '1B':
            p.proj = 1_000

    def comp(sum, a, b):
        return sum(b) >= sum(a) + 2

    # It appears to me closures do not work with ortools
    # (ex. passing p.team == t)
    custom_rules.append(
        CustomRule(
            # Given 1B in optimized lineup
            group_a=lambda p:
            p.pos == '1B' and p.team == 'ATL',

            # Ensure the stack is four players
            group_b=lambda p:
            '1B' not in p.pos and p.team == 'ATL',  # batters only

            comparison=comp,
        )
    )
    custom_rules.append(
        CustomRule(
            # Given 1B in optimized lineup
            group_a=lambda p:
            p.pos == '1B' and p.team == 'BOS',

            # Ensure the stack is four players
            group_b=lambda p:
            '1B' not in p.pos and p.team == 'BOS',  # batters only

            comparison=comp,
        )
    )

    settings = OptimizerSettings(
        custom_rules=custom_rules
    )

    roster = run(
        rule_set=rules.DK_MLB_RULE_SET,
        player_pool=player_pool,
        verbose=True,
        optimizer_settings=settings,
    )
    team_for_first = [p for p in roster.players if p.pos == '1B'][0].team
    total = len([p for p in roster.players
                 if p.team == team_for_first and 'P' not in p.pos])
    assert total > 3, f"{total} below 4"



from os import environ
from draftfast import rules
from draftfast.optimize import run_multi
from draftfast.csv_parse import salary_download, uploaders

"""
Script to create 20 NBA lineups for DraftKings.

Assumptions:
- Environment variable called "downloads" has path to downloads
- In downloads, salary, projection and player ID files exist with
  the filenames listed in the script.
"""
downloads = environ.get('downloads')
players = salary_download.generate_players_from_csvs(
    salary_file_location='{}/NBA_SALS.csv'.format(downloads),
    projection_file_location='{}/NBA_PROJECTIONS.csv'.format(downloads),
    game=rules.DRAFT_KINGS,
)

rosters, _ = run_multi(
    iterations=20,
    exposure_bounds=[
        {
            'name': 'Damian Lillard',
            'min': 0.3,
            'max': 0.5,
        },
        {
            'name': 'LaMarcus Aldridge',
            'min': 0.5,
            'max': 0.7,
        },
    ],
    rule_set=rules.DK_NBA_RULE_SET,
    player_pool=players,
    verbose=True,
)

uploader = uploaders.DraftKingsNBAUploader(
    pid_file='{}/NBA_PIDS.csv'.format(downloads),
)
uploader.write_rosters(rosters)



from draftfast.player_pool import add_pickem_contraints
from draftfast.pickem.pickem_orm import TieredLineup, TIERS
from draftfast.settings import PlayerPoolSettings
from draftfast.lineup_constraints import LineupConstraints
from draftfast.pickem.pickem_orm import TieredPlayer


def optimize(
    all_players: list,
    player_settings: PlayerPoolSettings = PlayerPoolSettings(),
    constraints: LineupConstraints = LineupConstraints()
):
    lineup_players = []
    all_players = list(filter(
        add_pickem_contraints(player_settings),
        all_players
    ))

    if constraints.has_group_constraints():
        raise NotImplementedError('Groups are not supported for pickem')

    for p in all_players:
        if constraints.is_banned(p.name):
            p.ban = True

    for t in TIERS:
        best = sorted(
            [p for p in all_players if p.tier == t and not p.ban],
            key=lambda p: p.proj,
            reverse=True,
        )[0]

        lineup_players.append(best)

    lineup = TieredLineup(lineup_players)

    for p in all_players:
        if constraints.is_locked(p.name):
            p.lock = True
            setattr(lineup, p.tier, p)

    return lineup


def _get_player(name, all_players) -> TieredPlayer:
    return next(
        p for p in all_players if p.name == name
    )



from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.orm import Player
from draftfast.settings import OptimizerSettings, CustomRule

mock_nba_pool = [
    Player(name='A1', cost=5500, proj=100, pos='PG'),
    Player(name='A2', cost=5500, proj=41, pos='PG'),
    Player(name='A100', cost=5500, proj=501, pos='PG'),
    Player(name='A101', cost=5500, proj=500, pos='PG'),
    Player(name='A11', cost=5500, proj=50, pos='PG'),
    Player(name='A3', cost=5500, proj=42, pos='SG'),
    Player(name='A4', cost=5500, proj=0, pos='SG'),
    Player(name='A5', cost=5500, proj=44, pos='SF'),
    Player(name='A6', cost=5500, proj=45, pos='SF'),
    Player(name='A7', cost=5500, proj=46, pos='PF'),
    Player(name='A8', cost=5500, proj=47, pos='PF'),
    Player(name='A9', cost=5500, proj=0, pos='C'),
    Player(name='A10', cost=5500, proj=49, pos='C'),
]

team = 0
for p in mock_nba_pool:
    team += 1
    if p.name in ['A101', 'A100', 'A10']:
        p.team = 'SomeTeam'
    else:
        p.team = str(team)


def test_if_one_then_two():
    # Base case
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
    )
    names = {p.name for p in roster.players}
    ntools.assert_equal(
        True,
        'A1' in names and 'A4' not in names and 'A9' not in names
    )
    settings = OptimizerSettings(
        custom_rules=[
            # Always play A1 with A9 and A4
            CustomRule(
                group_a=lambda p: p.name == 'A1',
                group_b=lambda p: p.name == 'A9' or p.name == 'A4'
            )
        ]
    )

    # TODO - add test where A1 is not in optimized lineup
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
        optimizer_settings=settings,
    )
    names = {p.name for p in roster.players}

    # Without this rule, A4 and A9 would never appear in the optimized
    # lineup. Both have a 0 point projection.
    ntools.assert_equal(
        True,
        'A1' in names and 'A9' in names and 'A4' in names
    )


def test_never_two():
    # Base case
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
    )
    names = {p.name for p in roster.players}
    ntools.assert_equal(
        True,
        'A101' in names and 'A100' in names
    )

    # Never play two players together
    settings = OptimizerSettings(
        custom_rules=[
          CustomRule(
              group_a=lambda p: p,
              group_b=lambda p: p.name == 'A100' or p.name == 'A101',
              comparison=lambda sum, a, b: sum(b) <= 1
          )
        ]
    )

    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
        optimizer_settings=settings,
    )
    names = {p.name for p in roster.players}

    # Without this rule, A4 and A9 would never appear in the optimized
    # lineup. Both have a 0 point projection.
    ntools.assert_equal(
        True,
        'A101' not in names and 'A100' in names
    )


def test_team_rules():
    # If two PGs on one team, play the C from same team
    settings = OptimizerSettings(
        custom_rules=[
            CustomRule(
                group_a=lambda p: p.pos == 'C' and p.team == 'SomeTeam',
                group_b=lambda p: p.pos == 'PG' and p.team == 'SomeTeam',
            )
        ]
    )
    roster = run(
        rule_set=rules.DK_NBA_RULE_SET,
        player_pool=mock_nba_pool,
        verbose=True,
        optimizer_settings=settings,
    )
    names = {p.name for p in roster.players}
    ntools.assert_equal(
        True,
        'A100' in names and 'A101' in names and 'A10' in names
    )



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

requires = [
    'numpy==1.15.4',
    'terminaltables==3.1.0',
    'ortools==7.6.7691',
]

setuptools.setup(
    name='draftfast',
    version='3.5.7',
    author='Ben Brostoff',
    author_email='ben.brostoff@gmail.com',
    description='A tool to automate and optimize DraftKings and FanDuel '
                'lineup construction.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/BenBrostoff/draftfast',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    install_requires=requires,
)



import os
from nose import tools as ntools
from draftfast.orm import Player
from draftfast.optimize import run
from draftfast import rules
from draftfast.csv_parse import salary_download as sd


def test_general_guard():
    pg = Player(name='A', cost=1, proj=1, pos='PG')
    ntools.assert_equal(pg.nba_general_position, 'G')
    sg = Player(name='A', cost=1, proj=1, pos='SG')
    ntools.assert_equal(sg.nba_general_position, 'G')


def test_general_forward():
    pg = Player(name='A', cost=1, proj=1, pos='SF')
    ntools.assert_equal(pg.nba_general_position, 'F')
    sg = Player(name='A', cost=1, proj=1, pos='PF')
    ntools.assert_equal(sg.nba_general_position, 'F')


def test_general_center():
    pg = Player(name='A', cost=1, proj=1, pos='C')
    ntools.assert_equal(pg.nba_general_position, 'C')


def test_optimize_with_general():
    def get_player_count_at_pos(roster, pos):
        return len([
            p for p in roster.players
            if p.nba_general_position == pos
        ])

    current_dir = os.path.dirname(os.path.abspath(__file__))
    players = sd.generate_players_from_csvs(
        game=rules.DRAFT_KINGS,
        salary_file_location='{}/data/dk-nba-salaries.csv'.format(current_dir)
    )

    rosters = []

    for _ in range(30):
        rosters.append(
            run(
                rule_set=rules.DK_NBA_RULE_SET,
                player_pool=players,
                verbose=True,
            )
        )

    # (Possibly) due to how ortools works,
    # same players are produced with different positions
    #
    # this does appear to be an ortools artifact. Since the lineup produced is
    # correct and optimal either way, i modified the test to account for all
    # possible constructions
    ntools.assert_equal(rosters[0].projected(), 279.53)

    for i in range(1, 30):
        print(i)

        ntools.assert_equal(rosters[i], rosters[0])

        ntools.assert_equal(rosters[i].projected(), 279.53)
        ntools.assert_true(get_player_count_at_pos(rosters[i], 'G') in [3, 4])
        ntools.assert_true(get_player_count_at_pos(rosters[i], 'F') in [3, 4])
        ntools.assert_true(get_player_count_at_pos(rosters[i], 'C') in [1, 2])



import random
from copy import deepcopy
from typing import List
from draftfast import player_pool as pool
from draftfast.orm import RosterSelect, Roster
from draftfast.optimizer import Optimizer
from draftfast.exposure import check_exposure, \
    get_exposure_table, get_exposure_matrix, get_exposure_args
from draftfast.rules import RuleSet
from draftfast.settings import PlayerPoolSettings, OptimizerSettings
from draftfast.lineup_constraints import LineupConstraints


def run(rule_set: RuleSet,
        player_pool: list,
        constraints: LineupConstraints = LineupConstraints(),
        optimizer_settings: OptimizerSettings = OptimizerSettings(),
        player_settings: PlayerPoolSettings = PlayerPoolSettings(),
        exposure_dict: dict = dict(),
        roster_gen: Roster = None,
        verbose=False) -> Roster:
    players = player_pool
    if player_settings.exist() or constraints.exist():
        players = pool.filter_pool(
            deepcopy(player_pool),
            player_settings,
        )

    if not isinstance(rule_set, RuleSet):
        raise Exception("RuleSet not defined. Please refer to the docs")

    if rule_set.game_type == 'showdown':
        if optimizer_settings.no_offense_against_defense:
            print('WARNING:')
            print('no_offense_against_defense setting ignored for showdown')
            print('game types. Use no_defense_against_captain instead.')
            print()

    optimizer = Optimizer(
        players=players,
        rule_set=rule_set,
        settings=optimizer_settings,
        lineup_constraints=constraints,
        exposure_dict=exposure_dict,
    )

    variables = optimizer.variables

    if optimizer.solve():
        if roster_gen:
            roster = roster_gen()
        else:
            roster = RosterSelect().roster_gen(rule_set.league)

        for i, player in enumerate(players):
            if variables[i].solution_value() == 1:
                roster.add_player(player)

        if verbose:
            print('Optimal roster for: {}'.format(rule_set.league))
            print(roster)

        return roster

    if verbose:
        print(
            '''
No solution found.
Try adjusting your query by taking away constraints.

OPTIMIZER CONSTRAINTS:

{}

LINEUP CONSTRAINTS:

{}

PLAYER POOL SETTINGS:

{}

PLAYER COUNT: {}
        '''.format(
                optimizer_settings,
                constraints,
                player_settings,
                len(players or [])
            )
        )
    return None


def run_multi(
    iterations: int,
    rule_set: RuleSet,
    player_pool: list,
    constraints: LineupConstraints = LineupConstraints(),
    player_settings: PlayerPoolSettings = PlayerPoolSettings(),
    optimizer_settings: OptimizerSettings = OptimizerSettings(),
    verbose=False,
    exposure_bounds: List[dict] = list(),
    exposure_random_seed=None,
) -> [List[Roster], list]:

    if not isinstance(rule_set, RuleSet):
        raise Exception("RuleSet not defined. Please refer to the docs")

    # set the random seed globally for random lineup exposure
    random.seed(exposure_random_seed)

    rosters = []
    for _ in range(0, iterations):
        exposure_dict = get_exposure_args(
            existing_rosters=optimizer_settings.existing_rosters,
            exposure_bounds=exposure_bounds,
            n=iterations,
            use_random=bool(exposure_random_seed),
            random_seed=exposure_random_seed,
        )

        roster = run(
            rule_set=rule_set,
            player_pool=player_pool,
            optimizer_settings=optimizer_settings,
            player_settings=player_settings,
            exposure_dict=exposure_dict,
            constraints=constraints,
            verbose=verbose,
        )
        if roster:
            optimizer_settings.existing_rosters += [roster]

        if roster:
            rosters.append(roster)
        else:
            break

        # clear ban/lock to reset exposure between iterations
        reset_player_ban_lock(player_pool)

    exposure_diffs = {}

    if rosters and verbose:
        print(get_exposure_table(rosters, exposure_bounds))
        print()
        print(get_exposure_matrix(rosters))
        print()

        exposure_diffs = check_exposure(rosters, exposure_bounds)
        for n, d in exposure_diffs.items():
            if d < 0:
                print('{} is UNDER exposure by {} lineups'.format(n, d))
            else:
                print('{} is OVER exposure by {} lineups'.format(n, d))

    return rosters, exposure_diffs


def reset_player_ban_lock(player_pool):
    for p in player_pool:
        p.ban = False
        p.lock = False



from nose import tools as ntools
from draftfast.optimize import run
from draftfast import rules
from draftfast.orm import Player
from draftfast.settings import OptimizerSettings
from draftfast.showdown.orm import ShowdownPlayer
from draftfast.lineup_constraints import LineupConstraints


def _build_mock_player_pool():
    player_pool = [
        Player(name='A1', cost=5500, proj=100, pos='QB',
               team='X', matchup='X@Y'),
        Player(name='A2', cost=5500, proj=41, pos='QB',
               team='X', matchup='X@Y'),
        Player(name='A11', cost=5500, proj=50, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A3', cost=5500, proj=42, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A4', cost=5500, proj=43, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A5', cost=5500, proj=44, pos='WR',
               team='X', matchup='X@Y'),
        Player(name='A6', cost=5500, proj=45, pos='RB',
               team='X', matchup='X@Y'),
        Player(name='A7', cost=5500, proj=46, pos='RB',
               team='X', matchup='X@Y'),
        Player(name='A8', cost=5500, proj=47, pos='RB',
               team='X', matchup='X@Y'),
        Player(name='A9', cost=5500, proj=48, pos='TE',
               team='X', matchup='X@Y'),
        Player(name='A10', cost=5500, proj=49, pos='TE',
               team='X', matchup='X@Y'),
        Player(name='A12', cost=5500, proj=51, pos='DST',
               team='X', matchup='X@Y'),
        Player(name='A14', cost=5500, proj=40, pos='QB',
               team='Y', matchup='X@Y'),
        Player(name='A21', cost=5500, proj=41, pos='QB',
               team='Y', matchup='X@Y'),
        Player(name='A11', cost=5500, proj=50, pos='WR',
               team='Y', matchup='X@Y'),
        Player(name='A31', cost=5500, proj=42, pos='WR',
               team='Y', matchup='X@Y'),
        Player(name='A41', cost=5500, proj=43, pos='WR',
               team='Y', matchup='X@Y'),
        Player(name='A51', cost=5500, proj=54, pos='WR',
               team='Y', matchup='X@Y'),
        Player(name='A61', cost=5500, proj=45, pos='RB',
               team='Y', matchup='X@Y'),
        Player(name='A71', cost=5500, proj=56, pos='RB',
               team='Y', matchup='X@Y'),
        Player(name='A81', cost=5500, proj=47, pos='RB',
               team='Y', matchup='X@Y'),
        Player(name='A91', cost=5500, proj=48, pos='TE',
               team='Y', matchup='X@Y'),
        Player(name='A110', cost=5500, proj=49, pos='TE',
               team='Y', matchup='X@Y'),
        Player(name='A112', cost=5500, proj=60, pos='DST',
               team='Y', matchup='X@Y'),
    ]

    def capt_boost(p):
        return Player(
            name=p.name,
            team=p.team,
            matchup=p.matchup,
            pos=p.pos,
            cost=p.cost * 1.5,
            proj=p.proj * 1.5,
        )

    captain_pool = [capt_boost(p) for p in player_pool]

    mock_dk_pool = []
    for p in player_pool:
        mock_dk_pool.append(ShowdownPlayer(p))
    for p in captain_pool:
        mock_dk_pool.append(ShowdownPlayer(p, captain=True))

    return mock_dk_pool


def test_nfl_dk_showdown_mock():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
        ),
        verbose=True
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 421)


def test_nfl_showdown_no_def_against_capt():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
            no_defense_against_captain=True,
        ),
        verbose=True
    )

    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 408.0)
    for p in roster.players:
        ntools.assert_not_equal(p.name, 'A112')


def test_nfl_showdown_lock_general():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
            no_defense_against_captain=True,
        ),
        constraints=LineupConstraints(
            locked=['A14'],
        ),
        verbose=True
    )
    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 399.0)
    ntools.assert_true('A14' in [x.name for x in roster.players])


def test_nfl_showdown_lock_captain():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
            no_defense_against_captain=True,
        ),
        constraints=LineupConstraints(
            position_locked=['A2 CPT X'],
        ),
        verbose=True
    )
    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 370.5)
    cpt = [x for x in roster.players if x.pos == 'CPT'][0]
    ntools.assert_equal('A2', cpt.name)


def test_nfl_showdown_lock_flex():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
            no_defense_against_captain=True,
        ),
        constraints=LineupConstraints(
            position_locked=['A1 FLEX X'],
        ),
        verbose=True
    )
    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 386.0)
    flex = [
        x for x in roster.players
        if x.pos == 'FLEX'
        and x.name == 'A1'
    ][0]
    ntools.assert_equal('A1', flex.name)


def test_nfl_showdown_ban_general():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
            no_defense_against_captain=True,
        ),
        constraints=LineupConstraints(
            banned=['A1'],
        ),
        verbose=True
    )
    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 334.0)
    ntools.assert_true('A1' not in [x.name for x in roster.players])


def test_nfl_showdown_ban_specific():
    mock_dk_pool = _build_mock_player_pool()

    roster = run(
        rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
        player_pool=mock_dk_pool,
        optimizer_settings=OptimizerSettings(
            showdown_teams=('X', 'Y'),
            no_defense_against_captain=True,
        ),
        constraints=LineupConstraints(
            position_banned=['A1 CPT X'],
        ),
        verbose=True
    )
    ntools.assert_not_equal(roster, None)
    ntools.assert_equal(roster.projected(), 386.0)
    flex = [
        x for x in roster.players
        if x.pos == 'FLEX'
        and x.name == 'A1'
    ][0]
    ntools.assert_equal('A1', flex.name)



from draftfast import rules
from draftfast.optimize import run
from draftfast.orm import Player
from draftfast.showdown.orm import ShowdownPlayer

player_pool = [
    Player(name='A1', cost=5500, proj=100, pos='QB',
           team='X', matchup='X@Y'),
    Player(name='A2', cost=5500, proj=41, pos='QB',
           team='X', matchup='X@Y'),
    Player(name='A11', cost=5500, proj=50, pos='WR',
           team='X', matchup='X@Y'),
    Player(name='A3', cost=5500, proj=42, pos='WR',
           team='X', matchup='X@Y'),
    Player(name='A4', cost=5500, proj=43, pos='WR',
           team='X', matchup='X@Y'),
    Player(name='A5', cost=5500, proj=44, pos='WR',
           team='X', matchup='X@Y'),
    Player(name='A6', cost=5500, proj=45, pos='RB',
           team='X', matchup='X@Y'),
    Player(name='A7', cost=5500, proj=46, pos='RB',
           team='X', matchup='X@Y'),
    Player(name='A8', cost=5500, proj=47, pos='RB',
           team='X', matchup='X@Y'),
    Player(name='A9', cost=5500, proj=48, pos='TE',
           team='X', matchup='X@Y'),
    Player(name='A10', cost=5500, proj=49, pos='TE',
           team='X', matchup='X@Y'),
    Player(name='A12', cost=5500, proj=51, pos='DST',
           team='X', matchup='X@Y'),
    Player(name='A14', cost=5500, proj=40, pos='QB',
           team='Y', matchup='X@Y'),
    Player(name='A21', cost=5500, proj=41, pos='QB',
           team='Y', matchup='X@Y'),
    Player(name='A11', cost=5500, proj=50, pos='WR',
           team='Y', matchup='X@Y'),
    Player(name='A31', cost=5500, proj=42, pos='WR',
           team='Y', matchup='X@Y'),
    Player(name='A41', cost=5500, proj=43, pos='WR',
           team='Y', matchup='X@Y'),
    Player(name='A51', cost=5500, proj=54, pos='WR',
           team='Y', matchup='X@Y'),
    Player(name='A61', cost=5500, proj=45, pos='RB',
           team='Y', matchup='X@Y'),
    Player(name='A71', cost=5500, proj=56, pos='RB',
           team='Y', matchup='X@Y'),
    Player(name='A81', cost=5500, proj=47, pos='RB',
           team='Y', matchup='X@Y'),
    Player(name='A91', cost=5500, proj=48, pos='TE',
           team='Y', matchup='X@Y'),
    Player(name='A110', cost=5500, proj=49, pos='TE',
           team='Y', matchup='X@Y'),
    Player(name='A112', cost=5500, proj=60, pos='DST',
           team='Y', matchup='X@Y'),
]


def capt_boost(p):
    return Player(
        name=p.name,
        team=p.team,
        matchup=p.matchup,
        pos=p.pos,
        cost=p.cost * 1.5,
        proj=p.proj * 1.5,
    )


captain_pool = [capt_boost(p) for p in player_pool]

mock_dk_pool = []
for p in player_pool:
    mock_dk_pool.append(ShowdownPlayer(p))
for p in captain_pool:
    mock_dk_pool.append(ShowdownPlayer(p, captain=True))

run(
    rule_set=rules.DK_NFL_SHOWDOWN_RULE_SET,
    player_pool=mock_dk_pool,
    verbose=True
)